from __future__ import print_function, unicode_literals
from ..utils import (
    check_function,
    action_function,
)
from ..run_result import CheckResult, ActionResult
from dcicutils import ff_utils
from dcicutils.env_utils import FF_PROD_BUCKET_ENV
import re
import requests
import json
import datetime
import time
import itertools
import random
from fuzzywuzzy import fuzz
import boto3
from .helpers import wrangler_utils
from collections import Counter

# use a random number to stagger checks
random_wait = 20


@check_function(cmp_to_last=False)
def workflow_run_has_deleted_input_file(connection, **kwargs):
    """Checks all wfrs that are not deleted, and have deleted input files
    There is an option to compare to the last, and only report new cases (cmp_to_last)
    The full output has 2 keys, because we report provenance wfrs but not run action on them
    problematic_provenance: stores uuid of deleted file, and the wfr that is not deleted
    problematic_wfr:        stores deleted file,  wfr to be deleted, and its downstream items (qcs and output files)
    """
    check = CheckResult(connection, 'workflow_run_has_deleted_input_file')
    check.status = "PASS"
    check.action = "patch_workflow_run_to_deleted"
    my_key = connection.ff_keys
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    # run the check
    search_query = 'search/?type=WorkflowRun&status!=deleted&input_files.value.status=deleted&limit=all'
    bad_wfrs = ff_utils.search_metadata(search_query, key=my_key)
    if kwargs.get('cmp_to_last', False):
        # filter out wfr uuids from last run if so desired
        prevchk = check.get_latest_result()
        if prevchk:
            prev_wfrs = prevchk.get('full_output', [])
            filtered = [b.get('uuid') for b in bad_wfrs if b.get('uuid') not in prev_wfrs]
            bad_wfrs = filtered
    if not bad_wfrs:
        check.summmary = check.description = "No live WorkflowRuns linked to deleted input Files"
        return check
    brief = str(len(bad_wfrs)) + " live WorkflowRuns linked to deleted input Files"
    # problematic_provenance stores uuid of deleted file, and the wfr that is not deleted
    # problematic_wfr stores deleted file,  wfr to be deleted, and its downstream items (qcs and output files)
    fulloutput = {'problematic_provenance': [], 'problematic_wfrs': []}
    no_of_items_to_delete = 0

    def fetch_wfr_associated(wfr_info):
        """Given wfr_uuid, find associated output files and qcs"""
        wfr_as_list = []
        wfr_as_list.append(wfr_info['uuid'])
        if wfr_info.get('output_files'):
            for o in wfr_info['output_files']:
                if o.get('value'):
                    wfr_as_list.append(o['value']['uuid'])
                if o.get('value_qc'):
                    wfr_as_list.append(o['value_qc']['uuid'])
        if wfr_info.get('output_quality_metrics'):
            for qc in wfr_info['output_quality_metrics']:
                if qc.get('value'):
                    wfr_as_list.append(qc['value']['uuid'])
        return list(set(wfr_as_list))

    for wfr in bad_wfrs:
        infiles = wfr.get('input_files', [])
        delfile = [f.get('value').get('uuid') for f in infiles if f.get('value').get('status') == 'deleted'][0]
        if wfr['display_title'].startswith('File Provenance Tracking'):
            fulloutput['problematic_provenance'].append([delfile, wfr['uuid']])
        else:
            del_list = fetch_wfr_associated(wfr)
            fulloutput['problematic_wfrs'].append([delfile, wfr['uuid'], del_list])
            no_of_items_to_delete += len(del_list)
    check.summary = "Live WorkflowRuns found linked to deleted Input Files"
    check.description = "{} live workflows were found linked to deleted input files - \
                         found {} items to delete, use action for cleanup".format(len(bad_wfrs), no_of_items_to_delete)
    if fulloutput.get('problematic_provenance'):
        brief += " ({} provenance tracking)"
    check.brief_output = brief
    check.full_output = fulloutput
    check.status = 'WARN'
    check.action_message = "Will attempt to patch %s workflow_runs with deleted inputs to status=deleted." % str(len(bad_wfrs))
    check.allow_action = True  # allows the action to be run
    return check


@action_function()
def patch_workflow_run_to_deleted(connection, **kwargs):
    action = ActionResult(connection, 'patch_workflow_run_to_deleted')
    check_res = action.get_associated_check_result(kwargs)
    action_logs = {'patch_failure': [], 'patch_success': []}
    my_key = connection.ff_keys
    for a_case in check_res['full_output']['problematic_wfrs']:
        wfruid = a_case[1]
        del_list = a_case[2]
        patch_data = {'status': 'deleted'}
        for delete_me in del_list:
            try:
                ff_utils.patch_metadata(patch_data, obj_id=delete_me, key=my_key)
            except Exception as e:
                acc_and_error = [delete_me, str(e)]
                action_logs['patch_failure'].append(acc_and_error)
            else:
                action_logs['patch_success'].append(wfruid + " - " + delete_me)
    action.output = action_logs
    action.status = 'DONE'
    if action_logs.get('patch_failure'):
        action.status = 'FAIL'
    return action


@check_function()
def item_counts_by_type(connection, **kwargs):
    def process_counts(count_str):
        # specifically formatted for FF health page
        ret = {}
        split_str = count_str.split()
        ret[split_str[0].strip(':')] = int(split_str[1])
        ret[split_str[2].strip(':')] = int(split_str[3])
        return ret

    check = CheckResult(connection, 'item_counts_by_type')
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    # run the check
    item_counts = {}
    warn_item_counts = {}
    req_location = ''.join([connection.ff_server, 'counts?format=json'])
    counts_res = ff_utils.authorized_request(req_location, auth=connection.ff_keys)
    if counts_res.status_code >= 400:
        check.status = 'ERROR'
        check.description = 'Error (bad status code %s) connecting to the counts endpoint at: %s.' % (counts_res.status_code, req_location)
        return check
    counts_json = json.loads(counts_res.text)
    for index in counts_json['db_es_compare']:
        counts = process_counts(counts_json['db_es_compare'][index])
        item_counts[index] = counts
        if counts['DB'] != counts['ES']:
            warn_item_counts[index] = counts
    # add ALL for total counts
    total_counts = process_counts(counts_json['db_es_total'])
    item_counts['ALL'] = total_counts
    # set fields, store result
    if not item_counts:
        check.status = 'FAIL'
        check.summary = check.description = 'Error on fourfront health page'
    elif warn_item_counts:
        check.status = 'WARN'
        check.summary = check.description = 'DB and ES item counts are not equal'
        check.brief_output = warn_item_counts
    else:
        check.status = 'PASS'
        check.summary = check.description = 'DB and ES item counts are equal'
    check.full_output = item_counts
    return check


@check_function()
def change_in_item_counts(connection, **kwargs):
    # use this check to get the comparison
    check = CheckResult(connection, 'change_in_item_counts')
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    counts_check = CheckResult(connection, 'item_counts_by_type')
    latest_check = counts_check.get_primary_result()
    # get_item_counts run closest to 10 mins
    prior_check = counts_check.get_closest_result(diff_hours=24)
    if not latest_check.get('full_output') or not prior_check.get('full_output'):
        check.status = 'ERROR'
        check.description = 'There are no counts_check results to run this check with.'
        return check
    diff_counts = {}
    # drill into full_output
    latest = latest_check['full_output']
    prior = prior_check['full_output']
    # get any keys that are in prior but not latest
    prior_unique = list(set(prior.keys()) - set(latest.keys()))
    for index in latest:
        if index == 'ALL':
            continue
        if index not in prior:
            diff_counts[index] = {'DB': latest[index]['DB'], 'ES': 0}
        else:
            diff_DB = latest[index]['DB'] - prior[index]['DB']
            if diff_DB != 0:
                diff_counts[index] = {'DB': diff_DB, 'ES': 0}
    for index in prior_unique:
        diff_counts[index] = {'DB': -1 * prior[index]['DB'], 'ES': 0}

    # now do a metadata search to make sure they match
    # date_created endpoints for the FF search
    # XXX: We should revisit if we really think this search is necessary. - will 3-26-2020
    to_date = datetime.datetime.strptime(latest_check['uuid'], "%Y-%m-%dT%H:%M:%S.%f").strftime('%Y-%m-%d+%H:%M')
    from_date = datetime.datetime.strptime(prior_check['uuid'], "%Y-%m-%dT%H:%M:%S.%f").strftime('%Y-%m-%d+%H:%M')
    # tracking items and ontology terms must be explicitly searched for
    search_query = ''.join(['search/?type=Item&type=OntologyTerm&type=TrackingItem',
                            '&frame=object&date_created.from=',
                            from_date, '&date_created.to=', to_date])
    search_resp = ff_utils.search_metadata(search_query, key=connection.ff_keys)
    # add deleted/replaced items
    search_query += '&status=deleted&status=replaced'
    search_resp.extend(ff_utils.search_metadata(search_query, key=connection.ff_keys))
    for res in search_resp:

        # Stick with given type name in CamelCase since this is now what we get on the counts page
        _type = res['@type'][0]
        _entry = diff_counts.get(_type)
        if not _entry:
            diff_counts[_type] = _entry = {'DB': 0, 'ES': 0}
        if _type in diff_counts:
            _entry['ES'] += 1

    check.ff_link = ''.join([connection.ff_server, 'search/?type=Item&',
                             'type=OntologyTerm&type=TrackingItem&date_created.from=',
                             from_date, '&date_created.to=', to_date])
    check.brief_output = diff_counts

    # total created items from diff counts (exclude any negative counts)
    total_counts_db = sum([diff_counts[coll]['DB'] for coll in diff_counts if diff_counts[coll]['DB'] >= 0])
    # see if we have negative counts
    # allow negative counts, but make note of, for the following types
    purged_types = ['TrackingItem', 'HiglassViewConfig']
    negative_types = [tp for tp in diff_counts if (diff_counts[tp]['DB'] < 0 and tp not in purged_types)]
    inconsistent_types = [tp for tp in diff_counts if (diff_counts[tp]['DB'] != diff_counts[tp]['ES'] and tp not in purged_types)]
    if negative_types:
        negative_str = ', '.join(negative_types)
        check.status = 'FAIL'
        check.summary = 'DB counts decreased in the past day for %s' % negative_str
        check.description = ('Positive numbers represent an increase in counts. '
                             'Some DB counts have decreased!')
    elif inconsistent_types:
        check.status = 'WARN'
        check.summary = 'Change in DB counts does not match search result for new items'
        check.description = ('Positive numbers represent an increase in counts. '
                             'The change in counts does not match search result for new items.')
    else:
        check.status = 'PASS'
        check.summary = 'There are %s new items in the past day' % total_counts_db
        check.description = check.summary + '. Positive numbers represent an increase in counts.'
    check.description += ' Excluded types: %s' % ', '.join(purged_types)
    return check


@check_function(file_type=None, status=None, file_format=None, search_add_on=None)
def identify_files_without_filesize(connection, **kwargs):
    check = CheckResult(connection, 'identify_files_without_filesize')
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    # must set this to be the function name of the action
    check.action = "patch_file_size"
    check.allow_action = True
    default_filetype = 'File'
    default_stati = 'released%20to%20project&status=released&status=uploaded&status=pre-release'
    filetype = kwargs.get('file_type') or default_filetype
    stati = 'status=' + (kwargs.get('status') or default_stati)
    search_query = 'search/?type={}&{}&frame=object&file_size=No value'.format(filetype, stati)
    ff = kwargs.get('file_format')
    if ff is not None:
        ff = '&file_format.file_format=' + ff
        search_query += ff
    addon = kwargs.get('search_add_on')
    if addon is not None:
        if not addon.startswith('&'):
            addon = '&' + addon
        search_query += addon
    problem_files = []
    file_hits = ff_utils.search_metadata(search_query, key=connection.ff_keys, page_limit=200)
    if not file_hits:
        check.allow_action = False
        check.summary = 'All files have file size'
        check.description = 'All files have file size'
        check.status = 'PASS'
        return check

    for hit in file_hits:
        hit_dict = {
            'accession': hit.get('accession'),
            'uuid': hit.get('uuid'),
            '@type': hit.get('@type'),
            'upload_key': hit.get('upload_key')
        }
        problem_files.append(hit_dict)
    check.brief_output = '{} files with no file size'.format(len(problem_files))
    check.full_output = problem_files
    check.status = 'WARN'
    check.summary = 'File metadata found without file_size'
    status_str = 'pre-release/released/released to project/uploaded'
    if kwargs.get('status'):
        status_str = kwargs.get('status')
    type_str = ''
    if kwargs.get('file_type'):
        type_str = kwargs.get('file_type') + ' '
    ff_str = ''
    if kwargs.get('file_format'):
        ff_str = kwargs.get('file_format') + ' '
    check.description = "{cnt} {type}{ff}files that are {st} don't have file_size.".format(
        cnt=len(problem_files), type=type_str, st=status_str, ff=ff_str)
    check.action_message = "Will attempt to patch file_size for %s files." % str(len(problem_files))
    check.allow_action = True  # allows the action to be run
    return check


@action_function()
def patch_file_size(connection, **kwargs):
    action = ActionResult(connection, 'patch_file_size')
    action_logs = {'s3_file_not_found': [], 'patch_failure': [], 'patch_success': []}
    # get the associated identify_files_without_filesize run result
    filesize_check_result = action.get_associated_check_result(kwargs)
    for hit in filesize_check_result.get('full_output', []):
        bucket = connection.ff_s3.outfile_bucket if 'FileProcessed' in hit['@type'] else connection.ff_s3.raw_file_bucket
        head_info = connection.ff_s3.does_key_exist(hit['upload_key'], bucket)
        if not head_info:
            action_logs['s3_file_not_found'].append(hit['accession'])
        else:
            patch_data = {'file_size': head_info['ContentLength']}
            try:
                ff_utils.patch_metadata(patch_data, obj_id=hit['uuid'], key=connection.ff_keys)
            except Exception as e:
                acc_and_error = '\n'.join([hit['accession'], str(e)])
                action_logs['patch_failure'].append(acc_and_error)
            else:
                action_logs['patch_success'].append(hit['accession'])
    action.status = 'DONE'
    action.output = action_logs
    return action


@check_function()
def validate_entrez_geneids(connection, **kwargs):
    ''' query ncbi to see if geneids are valid
    '''
    check = CheckResult(connection, 'validate_entrez_geneids')
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    problems = {}
    timeouts = 0
    search_query = 'search/?type=Gene&limit=all&field=geneid'
    genes = ff_utils.search_metadata(search_query, key=connection.ff_keys)
    if not genes:
        check.status = "FAIL"
        check.description = "Could not retrieve gene records from fourfront"
        return check
    geneids = [g.get('geneid') for g in genes]

    query = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=gene&id={id}"
    for gid in geneids:
        if timeouts > 5:
            check.status = "FAIL"
            check.description = "Too many ncbi timeouts. Maybe they're down."
            return check
        gquery = query.format(id=gid)
        # make 3 attempts to query gene at ncbi
        for count in range(3):
            resp = requests.get(gquery)
            if resp.status_code == 200:
                break
            if resp.status_code == 429:
                time.sleep(0.334)
                continue
            if count == 2:
                timeouts += 1
                problems[gid] = 'ncbi timeout'
        try:
            rtxt = resp.text
        except AttributeError:
            problems[gid] = 'empty response'
        else:
            if rtxt.startswith('Error'):
                problems[gid] = 'not a valid geneid'
    if problems:
        check.summary = "{} problematic entrez gene ids.".format(len(problems))
        check.brief_output = problems
        check.description = "Problematic Gene IDs found"
        check.status = "WARN"
    else:
        check.status = "PASS"
        check.description = "GENE IDs are all valid"
    return check


def semver2int(semver):
    v = [num for num in semver.lstrip('v').split('.')]
    for i in range(1,len(v)):
        if len(v[i]) == 1:
            v[i] = '0' + v[i]
    return float(''.join([v[0] + '.'] + v[1:]))


@check_function()
def check_for_ontology_updates(connection, **kwargs):
    '''
    Checks for updates in one of the three main ontologies that the 4DN data portal uses:
    EFO, UBERON, and OBI.
    EFO: checks github repo for new releases and compares release tag. Release tag is a
    semantic version number starting with 'v'.
    OBI: checks github repo for new releases and compares release tag. Release tag is a 'v'
    plus the release date.
    UBERON: github site doesn't have official 'releases' (and website isn't properly updated),
    so checks for commits that have a commit message containing 'new release'

    If version numbers to compare against aren't specified in the UI, it will use the ones
    from the previous primary check result.
    '''
    check = CheckResult(connection, 'check_for_ontology_updates')
    check.summary = ''
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    ontologies = ff_utils.search_metadata(
        'search/?type=Ontology&frame=object',
        key=connection.ff_keys
    )
    ontologies = [o for o in ontologies if o['ontology_prefix'] != '4DN']
    versions = {
        o['ontology_prefix']: {
            'current': o.get('current_ontology_version'),
            'needs_update': False
        } for o in ontologies
    }
    for o in ontologies:
        # UBERON needs different behavior
        if o['ontology_prefix'] == 'UBERON':
            uberon = requests.get('http://svn.code.sf.net/p/obo/svn/uberon/releases/')
            ub_release = uberon._content.decode('utf-8').split('</li>\n  <li>')[-1]
            versions['UBERON']['latest'] = ub_release[ub_release.index('>') + 1: ub_release.index('</a>')].rstrip('/')
        # instead of repos etc, check download url for ontology header to get version
        elif o.get('download_url'):
            owl = requests.get(o['download_url'], headers={"Range": "bytes=0-2000"})
            if owl.status_code == 404:
                versions[o['ontology_prefix']]['latest'] = 'WARN: 404 at download_url'
                check.summary = '404 at download_url'
                check.description = 'One or more ontologies has a download_url with a 404 error.'
                check.description += ' Please update ontology item or try again later.'
                check.status = 'WARN'
                continue
            if 'versionIRI' in owl.text:
                idx = owl.text.index('versionIRI')
                vline = owl.text[idx:idx+150]
                if 'releases'in vline:
                    vline = vline.split('/')
                    v = vline[vline.index('releases')+1]
                    versions[o['ontology_prefix']]['latest'] = v
                    continue
                else:
                    # looks for date string in versionIRI line
                    match = re.search('(20)?([0-9]{2})-[0-9]{2}-(20)?[0-9]{2}', vline)
                    if match:
                        v = match.group()
                        versions[o['ontology_prefix']]['latest'] = v
                        continue
            # SO removed version info from versionIRI, use date field instead
            if 'oboInOwl:date' in owl.text:
                idx = owl.text.index('>', owl.text.index('oboInOwl:date'))
                vline = owl.text[idx+1:owl.text.index('<', idx)]
                v = vline.split()[0]
                versions[o['ontology_prefix']]['latest'] = datetime.datetime.strptime(v, '%d:%m:%Y').strftime('%Y-%m-%d')
    check.brief_output = []
    for k, v in versions.items():
        if v.get('latest') and '404' in v['latest']:
            check.brief_output.append('{} - 404'.format(k))
        elif not v['current']:
            v['needs_update'] = True
            check.brief_output.append('{} needs update'.format(k))
        elif k == 'EFO' and semver2int(v['latest']) > semver2int(v['current']):
            v['needs_update'] = True
            check.brief_output.append('{} needs update'.format(k))
        elif k != 'EFO' and v['latest'] > v['current']:
            v['needs_update'] = True
            check.brief_output.append('{} needs update'.format(k))
        else:
            check.brief_output.append('{} OK'.format(k))
    check.full_output = versions
    num = ''.join(check.brief_output).count('update')
    if '404' not in check.summary:
        if num:
            check.summary = 'Ontology updates available'
            check.description = '{} ontologies need update'.format(num)
            check.status = 'WARN'
        else:
            check.summary = 'Ontologies up-to-date'
            check.description = 'No ontology updates needed'
            check.status = 'PASS'
        if num == 1 & versions['SO']['needs_update']:
            check.status = 'PASS'
    return check


@check_function()
def states_files_without_higlass_defaults(connection, **kwargs):
    check = CheckResult(connection, 'states_files_without_higlass_defaults')
    check.action = 'patch_states_files_higlass_defaults'
    check.full_output = {'to_add': {}, 'problematic_files': {}}
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    query = '/search/?file_type=chromatin states&type=File'
    res = ff_utils.search_metadata(query, key=connection.ff_keys)
    for re in res:
        if not re.get('higlass_defaults'):
            if not re.get('tags'):
                check.full_output['problematic_files'][re['accession']] = 'missing state tag'
            else:
                check.full_output['to_add'][re['accession']] = re["tags"]

    if check.full_output['to_add']:
        check.status = 'WARN'
        check.summary = 'Ready to patch higlass_defaults'
        check.description = 'Ready to patch higlass_defaults'
        check.allow_action = True
        check.action_message = 'Will patch higlass_defaults to %s items' % (len(check.full_output['to_add']))
    elif check.full_output['problematic_files']:
        check.status = 'WARN'
        check.summary = 'There are some files without states tags'
    else:
        check.status = 'PASS'
        check.summary = 'higlass_defaults are all set'
    return check


@action_function()
def patch_states_files_higlass_defaults(connection, **kwargs):
    action = ActionResult(connection, 'patch_states_files_higlass_defaults')
    check_res = action.get_associated_check_result(kwargs)
    action_logs = {'patch_success': [], 'patch_failure': [], 'missing_ref_file': []}
    total_patches = check_res['full_output']['to_add']

    s3 = boto3.resource('s3')
    bucket = s3.Bucket('elasticbeanstalk-%s-files' % FF_PROD_BUCKET_ENV)

    query = '/search/?type=FileReference'
    all_ref_files = ff_utils.search_metadata(query, key=connection.ff_keys)
    ref_files_tags = {}
    for ref_file in all_ref_files:
        if ref_file.get('tags'):
            for ref_file_tag in ref_file.get('tags'):
                if 'states' in ref_file_tag:
                    ref_files_tags[ref_file_tag] = {'uuid': ref_file['uuid'], 'accession': ref_file['accession']}

    for item, tag in total_patches.items():
        if ref_files_tags.get(tag[0]):
            buck_obj = ref_files_tags[tag[0]]['uuid'] + '/' + ref_files_tags[tag[0]]['accession'] + '.txt'
            obj = bucket.Object(buck_obj)
            body = obj.get()['Body'].read().decode('utf8')
            lines = body.split()
            states_colors = [item for num, item in enumerate(lines) if num % 2 != 0]
            patch = {'higlass_defaults': {'colorScale': states_colors}}
            try:
                ff_utils.patch_metadata(patch, item, key=connection.ff_keys)
            except Exception as e:
                action_logs['patch_failure'].append({item: str(e)})
            else:
                action_logs['patch_success'].append(item)
        else:
            action_logs['missing_ref_file'].append({item: 'missing rows_info reference file'})

    if action_logs['patch_failure'] or action_logs['missing_ref_file']:
        action.status = 'FAIL'
    else:
        action.status = 'DONE'
    action.output = action_logs
    return action


@check_function()
def check_suggested_enum_values(connection, **kwargs):
    """On our schemas we have have a list of suggested fields for
    suggested_enum tagged fields. A value that is not listed in this list
    can be accepted, and with this check we will find all values for
    each suggested enum field that is not in this list.
    There are 2 functions below:

    - find_suggested_enum
    This functions takes properties for a item type (taken from /profiles/)
    and goes field by field, looks for suggested enum lists, and is also
    recursive for taking care of sub-embedded objects (tagged as type=object).
    Additionally, it also takes ignored enum lists (enums which are not
    suggested, but are ignored in the subsequent search).

    * after running this function, we construct a search url for each field,
    where we exclude all values listed under suggested_enum (and ignored_enum)
    from the search: i.e. if it was FileProcessed field 'my_field' with options
    [val1, val2], url would be:
    /search/?type=FileProcessed&my_field!=val1&my_field!=val2&my_field!=No value

    - extract value
    Once we have the search result for a field, we disect it
    (again for subembbeded items or lists) to extract the field value, and =
    count occurences of each new value. (i.e. val3:10, val4:15)

    *deleted items are not considered by this check
    """
    check = CheckResult(connection, 'check_suggested_enum_values')
    # add random wait
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    # must set this to be the function name of the action
    check.action = "add_suggested_enum_values"

    def find_suggested_enum(properties, parent='', is_submember=False):
        """Filter schema propteries for fields with suggested enums.
        This functions takes properties for a item type (taken from /profiles/)
        and goes field by field, looks for suggested enum lists, and is also
        recursive for taking care of sub-embedded objects (tagged as
        type=object). It also looks fore ignored enum lists.
        """
        def is_subobject(field):
            if field.get('type') == 'object':
                return True
            try:
                return field['items']['type'] == 'object'
            except:
                return False

        def dotted_field_name(field_name, parent_name=None):
            if parent_name:
                return "%s.%s" % (parent_name, field_name)
            else:
                return field_name

        def get_field_type(field):
            field_type = field.get('type', '')
            if field_type == 'string':
                if field.get('linkTo', ''):
                    return "Item:" + field.get('linkTo')
                # if multiple objects are linked by "anyOf"
                if field.get('anyOf', ''):
                    links = list(filter(None, [d.get('linkTo', '') for d in field.get('anyOf')]))
                    if links:
                        return "Item:" + ' or '.join(links)
                # if not object return string
                return 'string'
            elif field_type == 'array':
                return 'array of ' + get_field_type(field.get('items'))
            return field_type

        fields = []
        for name, props in properties.items():
            options = []
            # focus on suggested_enum ones
            if 'suggested_enum' not in str(props):
                continue
            # skip calculated
            if props.get('calculatedProperty'):
                continue
            is_array = False
            if is_subobject(props) and name != 'attachment':
                is_array = get_field_type(props).startswith('array')
                obj_props = {}
                if is_array:
                    obj_props = props['items']['properties']
                else:
                    obj_props = props['properties']
                fields.extend(find_suggested_enum(obj_props, name, is_array))
            else:
                field_name = dotted_field_name(name, parent)
                field_type = get_field_type(props)
                # check props here
                if 'suggested_enum' in props:
                    options = props['suggested_enum']
                    if 'ignored_enum' in props:
                        options.extend(props['ignored_enum'])
                # if array of string with enum
                if is_submember or field_type.startswith('array'):
                    sub_props = props.get('items', '')
                    if 'suggested_enum' in sub_props:
                        options = sub_props['suggested_enum']
                        if 'ignored_enum' in sub_props:
                            options.extend(sub_props['ignored_enum'])
                fields.append((field_name, options))
        return(fields)

    def extract_value(field_name, item, options=[]):
        """Given a json, find the values for a given field.
        Once we have the search result for a field, we disect it
        (again for subembbeded items or lists) to extract the field value(s)
        """
        # let's exclude also empty new_values
        options.append('')
        new_vals = []
        if '.' in field_name:
            part1, part2 = field_name.split('.')
            val1 = item.get(part1)
            if isinstance(val1, list):
                for an_item in val1:
                    if an_item.get(part2):
                        new_vals.append(an_item[part2])
            else:
                if val1.get(part2):
                    new_vals.append(val1[part2])
        else:
            val1 = item.get(field_name)
            if val1:
                if isinstance(val1, list):
                    new_vals.extend(val1)
                else:
                    new_vals.append(val1)
        # are these linkTo items
        if new_vals:
            if isinstance(new_vals[0], dict):
                new_vals = [i['display_title'] for i in new_vals]
        new_vals = [i for i in new_vals if i not in options]
        return new_vals

    outputs = []
    # Get Schemas
    schemas = ff_utils.get_metadata('/profiles/', key=connection.ff_keys)
    sug_en_cases = {}
    for an_item_type in schemas:
        properties = schemas[an_item_type]['properties']
        sug_en_fields = find_suggested_enum(properties)
        if sug_en_fields:
            sug_en_cases[an_item_type] = sug_en_fields

    for item_type in sug_en_cases:
        for i in sug_en_cases[item_type]:
            extension = ""
            field_name = i[0]
            field_option = i[1]
            # create queries - we might need multiple since there is a url length limit
            # Experimental - limit seems to be between 5260-5340
            # all queries are appended by filter for No value
            character_limit = 2000
            extensions = []
            extension = ''
            for case in field_option:
                if len(extension) < character_limit:
                    extension += '&' + field_name + '!=' + case
                else:
                    # time to finalize, add no value
                    extension += '&' + field_name + '!=' + 'No value'
                    extensions.append(extension)
                    # reset extension
                    extension = '&' + field_name + '!=' + case
            # add the leftover extension - there should be always one
            if extension:
                extension += '&' + field_name + '!=' + 'No value'
                extensions.append(extension)

            # only return this field
            f_ex = '&field=' + field_name

            common_responses = []
            for an_ext in extensions:
                q = "/search/?type={it}{ex}{f_ex}".format(it=item_type, ex=an_ext, f_ex=f_ex)
                responses = ff_utils.search_metadata(q, connection.ff_keys)
                # if this is the first response, assign this as the first common response
                if not common_responses:
                    common_responses = responses
                # if it is the subsequent responses, filter the commons ones with the new requests (intersection)
                else:
                    filter_ids = [i['@id'] for i in responses]
                    common_responses = [i for i in common_responses if i['@id'] in filter_ids]
                # let's check if we depleted common_responses
                if not common_responses:
                    break

            odds = []
            for response in common_responses:
                odds.extend(extract_value(field_name, response, field_option))
            if len(odds) > 0:
                outputs.append(
                    {
                        'item_type': item_type,
                        'field': field_name,
                        'new_values': dict(Counter(odds))
                    })
    if not outputs:
        check.allow_action = False
        check.brief_output = []
        check.full_output = []
        check.status = 'PASS'
        check.summary = 'No new values for suggested enum fields'
        check.description = 'No new values for suggested enum fields'
    else:
        b_out = []
        for res in outputs:
            b_out.append(res['item_type'] + ': ' + res['field'])
        check.allow_action = False
        check.brief_output = b_out
        check.full_output = outputs
        check.status = 'WARN'
        check.summary = 'Suggested enum fields have new values'
        check.description = 'Suggested enum fields have new values'
    return check


@action_function()
def add_suggested_enum_values(connection, **kwargs):
    """No action is added yet, this is a placeholder for
    automated pr that adds the new values."""
    # TODO: for linkTo items, the current values are @ids, and might need a change
    action = ActionResult(connection, 'add_suggested_enum_values')
    action_logs = {}
    # check_result = action.get_associated_check_result(kwargs)
    action.status = 'DONE'
    action.output = action_logs
    return action


@check_function(days_back=30)
def check_external_references_uri(connection, **kwargs):
    '''
    Check if external_references.uri is missing while external_references.ref
    is present.
    '''
    check = CheckResult(connection, 'check_external_references_uri')

    days_back = kwargs.get('days_back')
    from_date_query, from_text = wrangler_utils.last_modified_from(days_back)

    search = ('search/?type=Item&external_references.ref%21=No+value' +
              '&field=external_references' + from_date_query)
    result = ff_utils.search_metadata(search, key=connection.ff_keys, is_generator=True)
    items = []
    for res in result:
        bad_refs = [er.get('ref') for er in res.get('external_references', []) if not er.get('uri')]
        if bad_refs:
            items.append({'@id': res['@id'], 'refs': bad_refs})
    names = [ref.split(':')[0] for item in items for ref in item['refs']]
    name_counts = [{na: names.count(na)} for na in set(names)]

    if items:
        check.status = 'WARN'
        check.summary = 'external_references.uri is missing'
        check.description = '%s items %sare missing uri' % (len(items), from_text)
    else:
        check.status = 'PASS'
        check.summary = 'All external_references uri are present'
        check.description = 'All dbxrefs %sare formatted properly' % from_text
    check.brief_output = name_counts
    check.full_output = items
    return check


def check_opf_lab_different_than_experiment(connection, **kwargs):
    '''
    Check if other processed files have lab (generating lab) that is different
    than the lab of that generated the experiment. In this case, the
    experimental lab needs to be added to the opf (contributing lab).
    '''
    check = CheckResult(connection, 'check_opf_lab_different_than_experiment')
    check.action = 'add_contributing_lab_opf'

    # check only recently modified files, to reduce the number of items
    days_back = kwargs.get('days_back')
    from_date_query, from_text = wrangler_utils.last_modified_from(days_back)

    search = ('search/?type=FileProcessed' +
              '&track_and_facet_info.experiment_bucket%21=No+value' +
              '&track_and_facet_info.experiment_bucket%21=processed+file' +
              '&field=experiment_sets&field=experiments' +
              '&field=lab&field=contributing_labs' + from_date_query)
    result = ff_utils.search_metadata(search, key=connection.ff_keys)

    opf = {'to_patch': [], 'problematic': []}
    exp_set_uuids = []  # Exp or ExpSet uuid list
    for res in result:
        if res.get('experiments'):
            if len(res['experiments']) != 1:  # this should not happen
                opf['problematic'].append({
                    '@id': res['@id'],
                    'experiments': [exp['uuid'] for exp in res['experiments']]})
                continue
            exp_or_set = res['experiments'][0]
        elif res.get('experiment_sets'):
            if len(res['experiment_sets']) != 1:  # this should not happen
                opf['problematic'].append({
                    '@id': res['@id'],
                    'experiment_sets': [es['uuid'] for es in res['experiment_sets']]})
                continue
            exp_or_set = res['experiment_sets'][0]
        else:  # this should not happen
            opf['problematic'].append({'@id': res['@id']})
            continue
        res['exp_set_uuid'] = exp_or_set['uuid']
        if res['exp_set_uuid'] not in exp_set_uuids:
            exp_set_uuids.append(res['exp_set_uuid'])

    # get lab of Exp/ExpSet
    result_exp_set = ff_utils.get_es_metadata(exp_set_uuids, sources=['uuid', 'properties.lab'], key=connection.ff_keys)
    uuid_2_lab = {}  # map file uuid to Exp/Set lab
    for item in result_exp_set:
        uuid_2_lab[item['uuid']] = item['properties']['lab']

    # evaluate contributing lab
    for res in result:
        if res['@id'] not in [pr['@id'] for pr in opf['problematic']]:
            contr_lab = []
            exp_set_lab = uuid_2_lab[res['exp_set_uuid']]
            if exp_set_lab == res['lab']['uuid']:
                continue
            elif res.get('contributing_labs'):
                contr_lab = [lab['uuid'] for lab in res['contributing_labs']]
                if exp_set_lab in contr_lab:
                    continue
            contr_lab.append(exp_set_lab)
            opf['to_patch'].append({
                '@id': res['@id'],
                'contributing_labs': contr_lab,
                'lab': res['lab']['display_title']})

    if opf['to_patch'] or opf['problematic']:
        check.status = 'WARN'
        check.summary = 'Supplementary files need attention'
        check.description = '%s files %sneed patching' % (len(opf['to_patch']), from_text)
        if opf['problematic']:
            check.description += ' and %s files have problems with experiments or sets' % len(opf['problematic'])
        if opf['to_patch']:
            check.allow_action = True
    else:
        check.status = 'PASS'
        check.summary = 'All supplementary files have correct contributing labs'
        check.description = 'All files %sare good' % from_text
    check.brief_output = {'to_patch': len(opf['to_patch']), 'problematic': len(opf['problematic'])}
    check.full_output = opf
    return check


@check_function()
def grouped_with_file_relation_consistency(connection, **kwargs):
    ''' Check if "grouped with" file relationships are reciprocal and complete.
        While other types of file relationships are automatically updated on
        the related file, "grouped with" ones need to be explicitly (manually)
        patched on the related file. This check ensures that there are no
        related files that lack the reciprocal relationship, or that lack some
        of the group relationships (for groups larger than 2 files).
    '''
    check = CheckResult(connection, 'grouped_with_file_relation_consistency')
    check.action = 'add_grouped_with_file_relation'
    search = 'search/?type=File&related_files.relationship_type=grouped+with&field=related_files'
    files = ff_utils.search_metadata(search, key=connection.ff_keys, is_generator=True)

    file2all = {}  # map all existing relations
    file2grp = {}  # map "group with" existing relations
    for f in files:
        for rel in f['related_files']:
            rel_type = rel['relationship_type']
            rel_file = rel['file']['@id']
            file2all.setdefault(f['@id'], []).append(
                {"relationship_type": rel_type, "file": rel_file})
            if rel_type == "grouped with":
                file2grp.setdefault(f['@id'], []).append(rel_file)

    # list groups of related items
    groups = []
    newgroups = [set(rel).union({file}) for file, rel in file2grp.items()]

    # Check if any pair of groups in the list has a common file (intersection).
    # In that case, they are parts of the same group: merge them.
    # Repeat until all groups are disjoint (not intersecting).
    while len(groups) != len(newgroups):
        groups, newgroups = newgroups, []
        for a_group in groups:
            for each_group in newgroups:
                if not a_group.isdisjoint(each_group):
                    each_group.update(a_group)
                    break
            else:
                newgroups.append(a_group)

    # find missing relations
    missing = {}
    for a_group in newgroups:
        pairs = [(a, b) for a in a_group for b in a_group if a != b]
        for (a_file, related) in pairs:
            if related not in file2grp.get(a_file, []):
                missing.setdefault(a_file, []).append(related)

    if missing:
        # add existing relations to patch related_files
        to_patch = {}
        for f, r in missing.items():
            to_patch[f] = file2all.get(f, [])
            to_patch[f].extend([{"relationship_type": "grouped with", "file": rel_f} for rel_f in r])
        check.brief_output = missing
        check.full_output = to_patch
        check.status = 'WARN'
        check.summary = 'File relationships are missing'
        check.description = "{} files are missing 'grouped with' relationships".format(len(missing))
        check.allow_action = True
        check.action_message = ("DO NOT RUN if relations need to be removed! "
            "This action will attempt to patch {} items by adding the missing 'grouped with' relations".format(len(to_patch)))
    else:
        check.status = 'PASS'
        check.summary = check.description = "All 'grouped with' file relationships are consistent"
    return check


@action_function()
def add_grouped_with_file_relation(connection, **kwargs):
    action = ActionResult(connection, 'add_grouped_with_file_relation')
    check_res = action.get_associated_check_result(kwargs)
    files_to_patch = check_res['full_output']
    action_logs = {'patch_success': [], 'patch_failure': []}
    for a_file, related_list in files_to_patch.items():
        patch_body = {"related_files": related_list}
        try:
            ff_utils.patch_metadata(patch_body, a_file, key=connection.ff_keys)
        except Exception as e:
            action_logs['patch_failure'].append({a_file: str(e)})
        else:
            action_logs['patch_success'].append(a_file)
    if action_logs['patch_failure']:
        action.status = 'FAIL'
    else:
        action.status = 'DONE'
    action.output = action_logs
    return action



from datetime import datetime


CHECK_DECO = 'check_function'
ACTION_DECO = 'action_function'


CHECK_TIMEOUT = 870  # in seconds. set to less than lambda limit (900 s)
POLL_INTERVAL = 10  # check child process every 10 seconds
LAMBDA_MAX_BODY_SIZE = 5500000  # 6Mb is the "real" threshold


ES_SEARCH_SIZE = 10000


# compare strings in both python 2 and python 3
# in other files, compare with vars.basestring
try:
    basestring = basestring
except NameError:
    basestring = str


def create_placeholder_check(check_name):
    """
    Creates a placeholder check for the given check_name to be rendered on the UI
    in the case that this check has not been run yet.
    """
    _uuid = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')
    placeholder = {
        'name': check_name,
        'uuid': _uuid,
        'kwargs': {'uuid': _uuid, 'primary': True},
        'status': 'PASS',  # so these show up green
        'summary': 'Check has not yet run',
        'description': 'If queued, this check will run with default arguments'
    }
    return placeholder



"""
Generate gitignored .chalice/config.json for deploy and then run deploy.
Takes on parameter for now: stage (either "dev" or "prod")
"""

from __future__ import print_function, unicode_literals
import os
import sys
import argparse
import json
import subprocess


class Deploy(object):
    
    CONFIG_BASE = {
      "stages": {
        "dev": {
          "api_gateway_stage": "api",
          "autogen_policy": False,
          "lambda_memory_size": 512,
          "lambda_timeout": 900,  # 15 mins in seconds
          "environment_variables": {
              "chalice_stage": "dev"
          }
        },
        "prod": {
          "api_gateway_stage": "api",
          "autogen_policy": False,
          "lambda_memory_size": 512,
          "lambda_timeout": 900,  # 15 mins in seconds
          "environment_variables": {
              "chalice_stage": "prod"
          }
        }
      },
      "version": "2.0",
      "app_name": "foursight-cgap",
      "layers": [  # required for Deployment Checks - Will 5/20/2020
          "arn:aws:lambda:us-east-1:553035198032:layer:git:11"
      ]
    }
    
    
    @classmethod
    def build_config_and_deploy(cls, stage):
        # key to de-encrypt access key
        s3_enc_secret = os.environ.get("S3_ENCRYPT_KEY")
        client_id = os.environ.get("CLIENT_ID")
        client_secret = os.environ.get("CLIENT_SECRET")
        dev_secret = os.environ.get("DEV_SECRET")
        if not (s3_enc_secret and client_id and client_secret and dev_secret):
            print(''.join(['ERROR. You are missing one more more environment ',
                           'variables needed to deploy Foursight.\n',
                           'Need: S3_ENCRYPT_KEY, CLIENT_ID, CLIENT_SECRET, DEV_SECRET.'])
                  )
            sys.exit()
        for curr_stage in ['dev', 'prod']:
            cls.CONFIG_BASE['stages'][curr_stage]['environment_variables']['S3_ENCRYPT_KEY'] = s3_enc_secret
            cls.CONFIG_BASE['stages'][curr_stage]['environment_variables']['CLIENT_ID'] = client_id
            cls.CONFIG_BASE['stages'][curr_stage]['environment_variables']['CLIENT_SECRET'] = client_secret
            cls.CONFIG_BASE['stages'][curr_stage]['environment_variables']['DEV_SECRET'] = dev_secret
    
        file_dir, _ = os.path.split(os.path.abspath(__file__))
        filename = os.path.join(file_dir, '.chalice/config.json')
        print(''.join(['Writing: ', filename]))
        with open(filename, 'w') as config_file:
            config_file.write(json.dumps(cls.CONFIG_BASE))
        # export poetry into requirements
        subprocess.check_call(['poetry', 'export', '-f', 'requirements.txt', '--without-hashes', '-o', 'requirements.txt'])
        # actually deploy
        subprocess.call(['chalice', 'deploy', '--stage', stage])
    
    
def main():
    parser = argparse.ArgumentParser('chalice_deploy')
    parser.add_argument(
        "stage",
        type=str,
        choices=['dev', 'prod'],
        help="chalice deployment stage. Must be one of 'prod' or 'dev'")
    args = parser.parse_args()
    Deploy.build_config_and_deploy(args.stage)


if __name__ == '__main__':
    main()



import sys
import datetime
import boto3
import json
sys.path.append('..')
import app
from foursight_core.chalicelib.es_connection import ESConnection
from foursight_core.chalicelib.vars import FOURSIGHT_PREFIX

# XXX: To use this script, run 'python migration.py <env> <stage>' in the root
# directory of this repository.
# Set MIGRATE to false to trigger an s3 clean
# Set PRIMARY to true to keep primary results while doing a clean
# You can also do - 'python migration.py' whe MIGRATE is true to automatically
# migrate all checks from all environments into the appropriate ES index
# This script reall just refactors the functionality in RunResult.delete_results
# and calls the migration check in es_checks.py.

ENVS = ['cgap']
STAGES = ['dev', 'prod']
ONE_WEEK_AGO = datetime.datetime.utcnow() - datetime.timedelta(days=7)
MIGRATE = False # set this option based on what you want to do
PRIMARY = False # set to true if you want to keep primary results

def filename_to_datetime(key):
    try:
        name, tstamp = key.split('/')
        time, sec, _ = tstamp.split('.')
        return datetime.datetime.strptime('.'.join([time, sec]), '%Y-%m-%dT%H:%M:%S.%f')
    except:
        return None

def filt(k):
    t = filename_to_datetime(k)
    if t is None:
        return False
    else:
        return t < ONE_WEEK_AGO

def clean(env, stage):
    bucket_name = FOURSIGHT_PREFIX + '-' + stage + '-' + env
    bucket = boto3.resource('s3').Bucket(bucket_name)
    client = boto3.client('s3')
    keys = []
    for obj in bucket.objects.all():
        keys.append(obj.key)

    keys_to_delete = list(filter(filt, keys))

    # filter on primary
    def is_not_primary(key):
        try:
            obj = json.loads(client.get_object(Bucket=bucket_name, Key=key)['Body'].read())
            return not obj['kwargs'].get('primary')
        except:
            return False

    # if primary is set to true in file then we keep them
    if PRIMARY:
        keys_to_delete = list(filter(is_not_primary, keys_to_delete))

    print("Total keys: %s" % len(keys))
    print("Keys to delete: %s" % len(keys_to_delete))
    for i in range(0, len(keys_to_delete), 1000):
        fmt = {'Objects': [{'Key': key} for key in keys_to_delete[i:i+1000]]}
        client.delete_objects(Bucket=bucket_name, Delete=fmt)

def migrate(env, stage):
    index_name = FOURSIGHT_PREFIX + '-' + stage + '-' + env

    app.set_stage(stage)
    app.set_timeout(0)
    es = ESConnection(index=index_name, host=HOST)
    conn = app.init_connection(env)
    args = {'timeout': 1000000, 'check_name': None, 'called_by': None}
    migrate = app.run_check_or_action(conn, 'es_checks/migrate_checks_to_es', args)
    diff = app.run_check_or_action(conn, 'es_checks/elasticsearch_s3_count_diff', {})
    print(migrate)
    print(diff)

def main():
    if MIGRATE:
        if len(sys.argv) < 3:
            for env in ENVS:
                for stage in STAGES:
                    migrate(env, stage)
        else:
            migrate(sys.argv[1], sys.argv[2])
        exit(0)
    else:
        clean(sys.argv[1], sys.argv[2])

if __name__ == '__main__':
    main()



import sys
import datetime
import boto3
import json
import argparse
sys.path.append('..')
import app
from chalicelib.vars import DEV_ENV


EPILOG = __doc__
ENVS = ['cgap']
STAGES = ['dev', 'prod']


def setup_stage(stage):
    if not stage:
        app.set_stage('dev')
    else:
        if stage not in STAGES:
            print('Bad stage')
            exit(1)
        else:
            app.set_stage(stage)


def setup_env(env):
    if not env:
        connection = app.init_connection(DEV_ENV)
    else:
        if env not in ENVS:
            print('Bad env')
            exit(1)
        try:
            connection = app.init_connection(env)
        except:
            print('Could not establish connection to env: %s' % env)
            exit(1)
    return connection


def main():
    parser = argparse.ArgumentParser(
        description='Tests a check/action',
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('check', help='full check name')
    parser.add_argument('--env', help='env name, %s by default' % DEV_ENV)
    parser.add_argument('--stage', help='stage, dev or prod, dev by default')
    args = parser.parse_args()

    setup_stage(args.stage)
    connection = setup_env(args.env)

    # run the check
    # XXX: Configure using kwargs?
    try:
        res = app.run_check_or_action(connection, args.check, {})
    except:
        print('Failed to execute check name: %s' % args.check)
        exit(1)
    print(res)

if __name__ == '__main__':
    main()



from __future__ import print_function, unicode_literals
from chalice import Response
from jinja2 import Environment, FileSystemLoader, select_autoescape
import json
import os
import jwt
import boto3
import datetime
import ast
import copy
import requests
import sys
from itertools import chain
from dateutil import tz
from base64 import b64decode
from dcicutils import ff_utils
from .s3_connection import S3Connection
from .fs_connection import FSConnection
from .utils import (
    basestring,
)
from .sqs_utils import (
    recover_message_and_propogate,
    delete_message_and_propogate,
    invoke_check_runner,
    get_sqs_queue,
    send_sqs_messages,
    get_sqs_attributes
)
from .vars import (
    FOURSIGHT_PREFIX,
    HOST
)
from .run_result import CheckResult, ActionResult
from .check_utils import CheckHandler
from .config import Config


class AppUtils(object):
    
    # these must be overwritten in inherited classes
    prefix = FOURSIGHT_PREFIX
    FAVICON = 'https://cgap.hms.harvard.edu/static/img/favicon-fs.ico'
    Config = Config
    CheckHandler = CheckHandler
    CheckResult = CheckResult
    ActionResult = ActionResult
    host = HOST

    TRIM_ERR_OUTPUT = 'Output too large to provide on main page - see check result directly'

    @classmethod
    def jin_env(cls):
        return Environment(
            loader=FileSystemLoader('chalicelib/templates'),
            autoescape=select_autoescape(['html', 'xml'])
        )

    @classmethod
    def init_environments(cls, env='all', envs=None):
        """
        Generate environment information from the envs bucket in s3.
        Returns a dictionary keyed by environment name with value of a sub-dict
        with the fields needed to initiate a connection.
    
        :param env: allows you to specify a single env to be initialized
        :param envs: allows you to specify multiple envs to be initialized
        """
        stage = cls.Config.get_stage_info()['stage']
        s3_connection = S3Connection(cls.prefix + '-envs')
        env_keys = s3_connection.list_all_keys()
        print("env_keys=%s from bucket %s" % (env_keys, cls.prefix + '-envs'))
        environments = {}
        if env != 'all':
            if env in env_keys:
                env_keys = [env]
            else:
                return {}  # provided env is not in s3
        elif envs is not None:
            env_keys = envs
        for env_key in env_keys:
            env_res = s3_connection.get_object(env_key)
            print("env_res for env_key %s from s3_connection = %s" % (env_key, str(env_res)))
            # check that the keys we need are in the object
            if isinstance(env_res, dict) and {'fourfront', 'es'} <= set(env_res):
                env_entry = {
                    'fourfront': env_res['fourfront'],
                    'es': env_res['es'],
                    'ff_env': env_res.get('ff_env', ''.join(['fourfront-', env_key])),
                    'bucket': ''.join([cls.prefix + '-', stage, '-', env_key])
                }
                environments[env_key] = env_entry
        return environments
    
    @classmethod
    def init_connection(cls, environ, _environments=None):
        """
        Initialize the fourfront/s3 connection using the FSConnection object
        and the given environment.
        Returns an FSConnection object or raises an error.
        """
        error_res = {}
        environments = cls.init_environments(environ) if _environments is None else _environments
        print("environments = %s" % str(environments))
        # if still not there, return an error
        if environ not in environments:
            error_res = {
                'status': 'error',
                'description': 'environment %s is not valid!' % environ,
                'environment': environ,
                'checks': {}
            }
            raise Exception(str(error_res))
        connection = FSConnection(environ, environments[environ], host=cls.host)
        return connection
    
    @classmethod
    def init_response(cls, environ):
        """
        Generalized function to init response given an environment
        """
        response = Response('Foursight response')
        try:
            connection = cls.init_connection(environ)
        except Exception as e:
            connection = None
            response.body = str(e)
            response.status_code = 400
        return connection, response
    
    @classmethod
    def check_authorization(cls, request_dict, env=None):
        """
        Manual authorization, since the builtin chalice @app.authorizer() was not
        working for me and was limited by a requirement that the authorization
        be in a token. Check the cookies of the request for jwtToken using utils
    
        Take in a dictionary format of the request (app.current_request) so we
        can test this.
        """
        # first check the Authorization header
        dev_auth = request_dict.get('headers', {}).get('authorization')
        # grant admin if dev_auth equals secret value
        if dev_auth and dev_auth == os.environ.get('DEV_SECRET'):
            return True
        # if we're on localhost, automatically grant authorization
        # this looks bad but isn't because request authentication will
        # still fail if local keys are not configured
        src_ip = request_dict.get('context', {}).get('identity', {}).get('sourceIp', '')
        if src_ip == '127.0.0.1':
            return True
        token = cls.get_jwt(request_dict)
        auth0_client = os.environ.get('CLIENT_ID', None)
        auth0_secret = os.environ.get('CLIENT_SECRET', None)
        if auth0_client and auth0_secret and token:
            try:
                if env is None:
                    return False  # we have no env to check auth
                # leeway accounts for clock drift between us and auth0
                payload = jwt.decode(token, b64decode(auth0_secret, '-_'), audience=auth0_client, leeway=30)
                for env_info in cls.init_environments(env).values():
                    user_res = ff_utils.get_metadata('users/' + payload.get('email').lower(),
                                                ff_env=env_info['ff_env'], add_on='frame=object')
                    if not ('admin' in user_res['groups'] and payload.get('email_verified')):
                        # if unauthorized for one, unauthorized for all
                        return False
                return True
            except:
                pass
        return False
    
    @classmethod
    def get_jwt(cls, request_dict):
        """
        Simple function to extract a jwt from a request that has already been
        dict-transformed
        """
        cookies = request_dict.get('headers', {}).get('cookie')
        cookie_dict = {}
        if cookies:
            for cookie in cookies.split(';'):
                cookie_split = cookie.strip().split('=')
                if len(cookie_split) == 2:
                    cookie_dict[cookie_split[0]] = cookie_split[1]
        token = cookie_dict.get('jwtToken', None)
        return token
    
    @classmethod
    def get_favicon(cls):
        """
        Returns faviron
        """
        return FAVICON  # want full HTTPS, so hard-coded in
    
    @classmethod
    def get_domain_and_context(cls, request_dict):
        """
        Given a request that has already been dict-transformed, get the host
        and the url context for endpoints. Context will basically either be
        '/api/' or '/'
        """
        domain = request_dict.get('headers', {}).get('host')
        context = '/api/' if request_dict.get('context', {}).get('path', '').startswith('/api/') else '/'
        return domain, context
    
    @classmethod
    def forbidden_response(cls, context="/"):
        return Response(status_code=403,
                        body='Forbidden. Login on the %s page.' % (context + 'view/<environment>'))
    
    @classmethod
    def process_response(cls, response):
        """
        Does any final processing of a Foursight response before returning it. Right now, this includes:
        * Changing the response body if it is greater than 5.5 MB (Lambda body max is 6 MB)
        """
        if cls.get_size(response.body) > LAMBDA_MAX_BODY_SIZE:  # should be much faster than json.dumps
            response.body = 'Body size exceeded 6 MB maximum.'
            response.status_code = 413
        return response
    
    @classmethod
    def query_params_to_literals(cls, params):
        """
        Simple function to loop through the query params and convert them to
        bools/ints/floats other literals as applicable
        """
        to_delete = []
        for key, value in params.items():
            if not value:
                # handles empty strings
                to_delete.append(key)
                continue
            try:
                as_literal = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                as_literal = value
            params[key] = as_literal
        for key in to_delete:
            del params[key]
        return params
    
    @classmethod
    def get_size(cls, obj, seen=None):
        """ Recursively finds size of objects
            Taken directly from: https://goshippo.com/blog/measure-real-size-any-python-object/
        """
        size = sys.getsizeof(obj)
        if seen is None:
            seen = set()
        obj_id = id(obj)
        if obj_id in seen:
            return 0
        # Important mark as seen *before* entering recursion to gracefully handle
        # self-referential objects
        seen.add(obj_id)
        if isinstance(obj, dict):
            size += sum([cls.get_size(v, seen) for v in obj.values()])
            size += sum([cls.get_size(k, seen) for k in obj.keys()])
        elif hasattr(obj, '__dict__'):
            size += cls.get_size(obj.__dict__, seen)
        elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
            size += sum([cls.get_size(i, seen) for i in obj])
        return size
    
    @classmethod
    def trim_output(cls, output, max_size=100000):
        """ Uses the helper above with sys.getsizeof to determine the output size and remove it if it is too large.
            Instead of encoding as JSON as that is very slow.
    
        Old docstring below:
    
        AWS lambda has a maximum body response size of 6MB. Since results are currently delivered entirely
        in the body of the response, let's limit the size of the 'full_output', 'brief_output', and
        'admin_output' fields to 100 KB (see if this is a reasonable amount).
        Slice the dictionaries, lists, or string to achieve this.
        max_size input integer is in bites
    
        Takes in the non-json formatted version of the fields. For now, just use this for /view/.
        """
        # formatted = json.dumps(output, indent=4)
        # if len(formatted) > max_size:
        #     return ''.join([formatted[:max_size], '\n\n... Output truncated ...'])
        # else:
        #     return formatted
        size = cls.get_size(output)
        if size > max_size:
            return cls.TRIM_ERR_OUTPUT
        return output
    
    
    ##### ROUTE RUNNING FUNCTIONS #####
    
    @classmethod
    def view_run_check(cls, environ, check, params, context="/"):
        """
        Called from the view endpoint (or manually, I guess), this queues the given
        check for the given environment and redirects to the view_foursight result
        for the new check.
        Params are kwargs that are read from the url query_params; they will be
        added to the kwargs used to run the check.
    
        Args:
            environ (str): Foursight environment name
            check (str): check function name
            params (dict): kwargs to use for check
            context (str): string context to use for Foursight routing
    
        Returns:
            chalicelib.Response: redirect to future check landing page
        """
        # convert string query params to literals
        params = cls.query_params_to_literals(params)
        queued_uuid = cls.queue_check(environ, check, params)
        # redirect to view page with a 302 so it isn't cached
        resp_headers = {'Location': '/'.join([context + 'view', environ, check, queued_uuid])}
        return Response(status_code=302, body=json.dumps(resp_headers),
                        headers=resp_headers)
    
    @classmethod
    def view_run_action(cls, environ, action, params, context="/"):
        """
        Called from the view endpoint (or manually, I guess), this runs the given
        action for the given environment and refreshes the foursight view.
        Params are kwargs that are read from the url query_params; they will be
        added to the kwargs used to run the check.
    
        Args:
            environ (str): Foursight environment name
            action (str): action function name
            params (dict): kwargs to use for check
            context (str): string context to use for Foursight routing
    
        Returns:
            chalicelib.Response: redirect to check view that called this action
        """
        # convert string query params to literals
        params = cls.query_params_to_literals(params)
        queued_uuid = cls.queue_action(environ, action, params)
        # redirect to calling check view page with a 302 so it isn't cached
        if 'check_name' in params and 'called_by' in params:
            check_detail = '/'.join([params['check_name'], params['called_by']])
            resp_headers = {'Location': '/'.join([context + 'view', environ, check_detail])}
        else:
            # no check so cannot redirect
            act_path = '/'.join([context + 'checks', action, queued_uuid])
            return Response(
                body = {
                    'status': 'success',
                    'details': 'Action is queued. When finished, view at: %s' % act_path,
                    'environment': environ
                },
                status_code = 200
            )
        return Response(status_code=302, body=json.dumps(resp_headers),
                        headers=resp_headers)
    
    @classmethod
    def view_foursight(cls, environ, is_admin=False, domain="", context="/"):
        """
        View a template of all checks from the given environment(s).
        Environ may be 'all' or a specific FS environments separated by commas.
        With 'all', this function can be somewhat slow.
        Domain is the current FS domain, needed for Auth0 redirect.
        Context is the current context, usually "/api/" or "/"
        Returns a response with html content.
        Non-protected route
        """
        html_resp = Response('Foursight viewing suite')
        html_resp.headers = {'Content-Type': 'text/html'}
        requested_envs = [e.strip() for e in environ.split(',')]
        environments = cls.init_environments(envs=requested_envs)  # cached at start of page load
        total_envs = []
        servers = []
        view_envs = environments.keys() if environ == 'all' else [e.strip() for e in environ.split(',')]
        for this_environ in view_envs:
            try:
                if not is_admin:  # no view permissions for non-admins on CGAP
                    continue
                connection = cls.init_connection(this_environ, _environments=environments)
            except Exception:
                connection = None
            if connection:
                servers.append(connection.ff_server)
                grouped_results = cls.CheckHandler().get_grouped_check_results(connection)
                for group in grouped_results:
                    for title, result in group.items():
                        if title == '_name':
                            continue
                        elif title == '_statuses':
                            # convert counts to strings for jinja
                            for stat, val in group[title].items():
                                group[title][stat] = str(val)
                            continue
                        else:
                            group[title] = cls.process_view_result(connection, result, is_admin)
                total_envs.append({
                    'status': 'success',
                    'environment': this_environ,
                    'groups': grouped_results
                })
        # prioritize these environments
        env_order = ['data', 'staging', 'webdev', 'hotseat', 'cgap']
        total_envs = sorted(total_envs, key=lambda v: env_order.index(v['environment']) if v['environment'] in env_order else 9999)
        template = cls.jin_env().get_template('view_groups.html')
        # get queue information
        queue_attr = get_sqs_attributes(get_sqs_queue().url)
        running_checks = queue_attr.get('ApproximateNumberOfMessagesNotVisible')
        queued_checks = queue_attr.get('ApproximateNumberOfMessages')
        first_env_favicon = cls.get_favicon()
        html_resp.body = template.render(
            env=environ,
            view_envs=total_envs,
            stage=cls.Config.get_stage_info()['stage'],
            load_time=cls.get_load_time(),
            is_admin=is_admin,
            domain=domain,
            context=context,
            running_checks=running_checks,
            queued_checks=queued_checks,
            favicon=first_env_favicon
        )
        html_resp.status_code = 200
        return cls.process_response(html_resp)
    
    @classmethod
    def view_foursight_check(cls, environ, check, uuid, is_admin=False, domain="", context="/"):
        """
        View a formatted html response for a single check (environ, check, uuid)
        """
        html_resp = Response('Foursight viewing suite')
        html_resp.headers = {'Content-Type': 'text/html'}
        total_envs = []
        servers = []
        try:
            connection = cls.init_connection(environ)
        except Exception:
            connection = None
        if connection:
            servers.append(connection.ff_server)
            res_check = cls.CheckResult(connection, check)
            if res_check:
                data = res_check.get_result_by_uuid(uuid)
                if data is None:
                    # the check hasn't run. Return a placeholder view
                    data = {
                        'name': check,
                        'uuid': uuid,
                        'status': 'ERROR',  # in this case we just queued a check, so ERROR is ok
                        'summary': 'Check has not yet run',
                        'description': 'Check has not yet run'
                    }
                title = cls.CheckHandler().get_check_title_from_setup(check)
                processed_result = cls.process_view_result(connection, data, is_admin)
                total_envs.append({
                    'status': 'success',
                    'environment': environ,
                    'checks': {title: processed_result}
                })
        template = cls.jin_env().get_template('view_checks.html')
        queue_attr = get_sqs_attributes(get_sqs_queue().url)
        running_checks = queue_attr.get('ApproximateNumberOfMessagesNotVisible')
        queued_checks = queue_attr.get('ApproximateNumberOfMessages')
        first_env_favicon = cls.get_favicon()
        html_resp.body = template.render(
            env=environ,
            view_envs=total_envs,
            stage=cls.Config.get_stage_info()['stage'],
            load_time=cls.get_load_time(),
            is_admin=is_admin,
            domain=domain,
            context=context,
            running_checks=running_checks,
            queued_checks=queued_checks,
            favicon=first_env_favicon
        )
        html_resp.status_code = 200
        return cls.process_response(html_resp)
    
    @classmethod
    def get_load_time(cls):
        """
        Returns the current time in ET, formatted the same was process_view_result
        """
        ts_utc = datetime.datetime.utcnow().replace(microsecond=0)
        ts_utc = ts_utc.replace(tzinfo=tz.tzutc())
        # change timezone to EST (specific location needed for daylight savings)
        ts_local = ts_utc.astimezone(tz.gettz('America/New_York'))
        return ''.join([str(ts_local.date()), ' at ', str(ts_local.time()), ' (', str(ts_local.tzname()), ')'])
    
    @classmethod
    def process_view_result(cls, connection, res, is_admin):
        """
        Do some processing on the content of one check result (res arg, a dict)
        Processes timestamp string, trims output fields, and adds action info.
    
        For action info, if the check has an action, try to find the associated
        action by looking for '<check name>/action_records<check uuid>' object in
        s3. The contents will be the path to the action. If found, display as
        the "associated action" and disabled further runs of the action from the
        check; also edit the check summary to reflect that action has finished.
        Otherwise, allow runs of the action.
        For now, always show latest action as well.
        """
        # first check to see if res is just a string, meaning
        # the check didn't execute properly
        if not isinstance(res, dict):
            error_res = {
                'status': 'ERROR',
                'content': True,
                'title': 'Check System Error',
                'description': res,
                'uuid': 'Did not successfully run'
            }
            return error_res
        # this can be removed once uuid has been around long enough
        ts_utc = res['uuid'] if 'uuid' in res else res['timestamp']
        ts_utc = datetime.datetime.strptime(ts_utc, "%Y-%m-%dT%H:%M:%S.%f").replace(microsecond=0)
        ts_utc = ts_utc.replace(tzinfo=tz.tzutc())
        # change timezone to EST (specific location needed for daylight savings)
        ts_local = ts_utc.astimezone(tz.gettz('America/New_York'))
        proc_ts = ''.join([str(ts_local.date()), ' at ', str(ts_local.time())])
        res['local_time'] = proc_ts
        if res.get('brief_output'):
            res['brief_output'] = json.dumps(trim_output(res['brief_output']), indent=2)
        if res.get('full_output'):
            res['full_output'] = json.dumps(trim_output(res['full_output']), indent=2)
        # only return admin_output if an admin is logged in
        if res.get('admin_output') and is_admin:
            res['admin_output'] = json.dumps(trim_output(res['admin_output']), indent=2)
        else:
            res['admin_output'] = None
    
        # ### LOGIC FOR VIEWING ACTION ###
        # if this check has already run an action, display that. Otherwise, allow
        # action to be run.
        # For now also get the latest result for the checks action
        if res.get('action'):
            action = cls.ActionResult(connection, res.get('action'))
            if action:
                action_record_key = '/'.join([res['name'], 'action_records', res['uuid']])
                assc_action_key = connection.connections['s3'].get_object(action_record_key)
                if assc_action_key:
                    assc_action_key = assc_action_key.decode()  # in bytes
                    assc_action = connection.get_object(assc_action_key)
                    # If assc_action_key is written but assc_action is None, then
                    # it most likely means the action is still running
                    if assc_action is not None:
                        res['assc_action_status'] = assc_action['status']
                        res['assc_action'] = json.dumps(assc_action, indent=4)
                        # update check summary
                        if res.get('summary'):
                            res['summary'] = 'ACTION %s: %s' % (assc_action['status'], res['summary'])
                    else:
                        res['assc_action_status'] = 'PEND'
                        res['assc_action'] = 'Associated action has not finished.'
                        # update check summary
                        if res.get('summary'):
                            res['summary'] = 'ACTION PENDING: %s' % res['summary']
                    # don't allow the action to be run again from this check
                    del res['action']
                    res['allow_action'] = False
                elif res.get('allow_action') is True:
                    # if there is an action + allow action is set but the action has
                    # not yet run, display an icon status to signify this
                    res['assc_action_status'] = 'ready'
    
                # This used to try to get the latest result and only populate 'latest_action' if one exists.
                # Doing so makes the main page take 2-3x as long to load, so we won't be doing that anymore.
                res['action_history'] = res.get('action')  # = action name
    
            else:
                del res['action']
        return res
    
    @classmethod
    def view_foursight_history(cls, environ, check, start=0, limit=25, is_admin=False,
                               domain="", context="/"):
        """
        View a tabular format of the history of a given check or action (str name
        as the 'check' parameter) for the given environment. Results look like:
        status, kwargs.
        start controls where the first result is and limit controls how many
        results are retrieved (see get_foursight_history()).
        Returns html.
        """
        html_resp = Response('Foursight history view')
        html_resp.headers = {'Content-Type': 'text/html'}
        server = None
        try:
            connection = cls.init_connection(environ)
        except Exception:
            connection = None
        if connection:
            server = connection.ff_server
            history = cls.get_foursight_history(connection, check, start, limit)
            history_kwargs = list(set(chain.from_iterable([l[2] for l in history])))
        else:
            history, history_kwargs = [], []
        template = cls.jin_env().get_template('history.html')
        check_title = cls.CheckHandler().get_check_title_from_setup(check)
        page_title = ''.join(['History for ', check_title, ' (', environ, ')'])
        queue_attr = get_sqs_attributes(get_sqs_queue().url)
        running_checks = queue_attr.get('ApproximateNumberOfMessagesNotVisible')
        queued_checks = queue_attr.get('ApproximateNumberOfMessages')
        favicon = cls.get_favicon()
        html_resp.body = template.render(
            env=environ,
            check=check,
            load_time=cls.get_load_time(),
            history=history,
            history_kwargs=history_kwargs,
            res_start=start,
            res_limit=limit,
            res_actual=len(history),
            page_title=page_title,
            stage=cls.Config.get_stage_info()['stage'],
            is_admin=is_admin,
            domain=domain,
            context=context,
            running_checks=running_checks,
            queued_checks=queued_checks,
            favicon=favicon
        )
        html_resp.status_code = 200
        return cls.process_response(html_resp)

    @classmethod
    def get_foursight_history(cls, connection, check, start, limit):
        """
        Get a brief form of the historical results for a check, including
        UUID, status, kwargs. Limit the number of results recieved to 500, unless
        otherwise specified ('limit' arg). 'start' arg determines where the start
        of the results grabbed is, with idx = 0 being the most recent one.
    
        'check' may be a check or an action (string name)
        """
        # limit 'limit' param to 500
        limit = 500 if limit > 500 else limit
        result_obj = cls.CheckHandler.init_check_or_action_res(connection, check)
        if not result_obj:
            return []
        return result_obj.get_result_history(start, limit)
    
    @classmethod
    def run_get_check(cls, environ, check, uuid=None):
        """
        Loads a specific check or action result given an environment, check or
        action name, and uuid (all strings).
        If uuid is not provided, get the primary_result.
        """
        connection, response = cls.init_response(environ)
        if not connection:
            return response
        res_obj = cls.CheckHandler.init_check_or_action_res(connection, check)
        if not res_obj:
            response.body = {
                'status': 'error',
                'description': 'Not a valid check or action.'
            }
            response.status_code = 400
        else:
            if uuid:
                data = res_obj.get_result_by_uuid(uuid)
            else:
                data = res_obj.get_primary_result()
            response.body = {
                'status': 'success',
                'data': data
            }
            response.status_code = 200
        return cls.process_response(response)

    @classmethod
    def run_put_check(cls, environ, check, put_data):
        """
        Abstraction of put_check functionality to allow for testing outside of chalice
        framework. Returns a response object
        """
        connection, response = cls.init_response(environ)
        if not connection:
            return response
        if not isinstance(put_data, dict):
            response.body = {
                'status': 'error',
                'endpoint': 'put_check',
                'check': check,
                'description': ' '.join(['PUT request is malformed:', str(put_data)]),
                'environment': environ
            }
            response.status_code = 400
            return response
        put_uuid = put_data.get('uuid', datetime.datetime.utcnow().isoformat())
        putCheck = cls.CheckResult(connection, check, init_uuid=put_uuid)
        # set valid fields from the PUT body. should this be dynamic?
        # if status is not included, it will be set to ERROR
        for field in ['title', 'status', 'summary', 'description', 'brief_output', 'full_output', 'admin_output']:
            put_content = put_data.get(field)
            prev_content = getattr(putCheck, field, None)
            if put_content:
                # append attribute data for _output fields if there are pre-existing
                # values originating from an existing put_uuid
                if prev_content and field in ['full_output', 'brief_output', 'admin_output']:
                    # will be list, dict, or string. make sure they are same type
                    if isinstance(prev_content, dict) and isinstance(put_content, dict):
                        prev_content.update(put_content)
                    elif isinstance(prev_content, list) and isinstance(put_content, list):
                        prev_content.extend(put_content)
                    elif isinstance(prev_content, basestring) and isinstance(put_content, basestring):
                        prev_content = prev_content + put_content
                    else:
                        # cannot append, just update with new
                        prev_content = put_content
                    setattr(putCheck, field, prev_content)
                else:
                    setattr(putCheck, field, put_content)
        # set 'primary' kwarg so that the result is stored as 'latest'
        putCheck.kwargs = {'primary': True, 'uuid': put_uuid}
        stored = putCheck.store_result()
        response.body = {
            'status': 'success',
            'endpoint': 'put_check',
            'check': check,
            'updated_content': stored,
            'environment': environ
        }
        response.status_code = 200
        return cls.process_response(response)
    
    @classmethod
    def run_put_environment(cls, environ, env_data):
        """
        Abstraction of the functionality of put_environment without the current_request
        to allow for testing.
        """
        proc_environ = environ.split('-')[-1] if environ.startswith('fourfront-') else environ
        response = None
        if isinstance(env_data, dict) and {'fourfront', 'es'} <= set(env_data):
            ff_address = env_data['fourfront'] if env_data['fourfront'].endswith('/') else env_data['fourfront'] + '/'
            es_address = env_data['es'] if env_data['es'].endswith('/') else env_data['es'] + '/'
            ff_env = env_data['ff_env'] if 'ff_env' in env_data else ''.join(['fourfront-', proc_environ])
            env_entry = {
                'fourfront': ff_address,
                'es': es_address,
                'ff_env': ff_env
            }
            s3_connection = S3Connection(cls.prefix + '-envs')
            s3_connection.put_object(proc_environ, json.dumps(env_entry))
            stage = cls.Config.get_stage_info()['stage']
            s3_bucket = ''.join([cls.prefix + '-', stage, '-', proc_environ])
            bucket_res = s3_connection.create_bucket(s3_bucket)
            if not bucket_res:
                response = Response(
                    body = {
                        'status': 'error',
                        'description': ' '.join(['Could not create bucket:', s3_bucket]),
                        'environment': proc_environ
                    },
                    status_code = 500
                )
            else:
                # if not testing, queue checks with 'put_env' condition for the new env
                if 'test' not in cls.Config.get_stage_info()['queue_name']:
                    for sched in cls.get_schedule_names():
                        cls.queue_scheduled_checks(environ, sched, conditions=['put_env'])
                response = Response(
                    body = {
                        'status': 'success',
                        'description': ' '.join(['Succesfully made:', proc_environ]),
                        'environment': proc_environ
                    },
                    status_code = 200
                )
        else:
            response = Response(
                body = {
                    'status': 'error',
                    'description': 'Environment creation failed',
                    'body': env_data,
                    'environment': proc_environ
                },
                status_code = 400
            )
        return cls.process_response(response)
    
    @classmethod
    def run_get_environment(cls, environ):
        """
        Return config information about a given environment, or throw an error
        if it is not valid.
        """
        environments = cls.init_environments()
        if environ in environments:
            response = Response(
                body = {
                    'status': 'success',
                    'details': environments[environ],
                    'environment': environ
                },
                status_code = 200
            )
        else:
            response = Response(
                body = {
                    'status': 'error',
                    'description': 'Invalid environment provided. Should be one of: %s' % (str(list(environments.keys()))),
                    'environment': environ
                },
                status_code = 400
            )
        return cls.process_response(response)
    
    @classmethod
    def run_delete_environment(cls, environ, bucket=FOURSIGHT_PREFIX + '-envs'):
        """
        Removes the environ entry from the Foursight envs bucket. This effectively de-schedules all checks
        but does not remove any data.
        """
        s3_connection = S3Connection(bucket)
        s3_resp = s3_connection.delete_keys([environ])
        keys_deleted = s3_resp['Deleted']
        if not keys_deleted:
            response = Response(
                body = {
                    'status': 'error',
                    'description': 'Unable to comply with request',
                    'environment': environ
                },
                status_code = 400
            )
        else:
            our_key = keys_deleted[0]  # since we only passed one key to be deleted, the response will be a length 1 list
            if our_key['Key'] != environ:
                response = Response(
                    body = {
                        'status': 'error',
                        'description': 'An error occurred during environment deletion, please check S3 directly',
                        'environment': environ
                    },
                    status_code = 400
                )
            else:  # we were successful
                response = Response(
                    body = {
                        'status': 'success',
                        'details': 'Successfully deleted environment %s' % environ,
                        'environment': environ
                    },
                    status_code = 200
                )
        return cls.process_response(response)
    
    
    
    ##### QUEUE / CHECK RUNNER FUNCTIONS #####
    
    @classmethod
    def queue_scheduled_checks(cls, sched_environ, schedule_name, conditions=None):
        """
        Given a str environment and schedule name, add the check info to the
        existing queue (or creates a new one if there is none). Then initiates 4
        check runners that are linked to the queue that are self-propogating.
    
        If sched_environ == 'all', then loop through all in Config.list_environments()
    
        Run with schedule_name = None to skip adding the check group to the queue
        and just initiate the check runners.
    
        Can optionally provide a list of conditions that will be used as used to
        filter the checks to schedule based on the 'conditions' list in check_setup
    
        Args:
            sched_environ (str): Foursight environment name to schedule on
            schedule_name (str): schedule name from check_setup / app
            conditions (list): optional list of one or more conditions to filter by
    
        Returns:
            dict: runner input of queued messages, used for testing
        """
        queue = get_sqs_queue()
        if schedule_name is not None:
            if sched_environ != 'all' and sched_environ not in Confnig.list_environments():
                print('-RUN-> %s is not a valid environment. Cannot queue.' % sched_environ)
                return
            sched_environs = cls.Config.list_environments() if sched_environ == 'all' else [sched_environ]
            check_schedule = cls.CheckHandler().get_check_schedule(schedule_name, conditions)
            if not check_schedule:
                print('-RUN-> %s is not a valid schedule. Cannot queue.' % schedule_name)
                return
            for environ in sched_environs:
                # add the run info from 'all' as well as this specific environ
                check_vals = copy.copy(check_schedule.get('all', []))
                check_vals.extend(check_schedule.get(environ, []))
                send_sqs_messages(queue, environ, check_vals)
        runner_input = {'sqs_url': queue.url}
        for n in range(4): # number of parallel runners to kick off
            invoke_check_runner(runner_input)
        return runner_input # for testing purposes
    
    @classmethod
    def queue_check(cls, environ, check, params={}, deps=[], uuid=None):
        """
        Queue a single check, given by check function name, with given parameters
        and dependencies (both optional). Also optionally pass in a uuid, which
        will be used for the run if provided
    
        Args:
            environ (str): Foursight environment name
            check (str): check function name
            params (dict): kwargs to use for check. Defaults to {}
            deps (list): list of dependencies for the check. Defaults to []
            uuid (str): optional uuid to pass to the run. Defaults to None
    
        Returns:
            str: uuid of the queued check (from send_single_to_queue)
        """
        check_str = cls.CheckHandler.get_check_strings(check)
        if not check_str:
            error_res = {
                'status': 'error',
                'description': 'could not find check %s' % check,
                'environment': environ,
                'checks': {}
            }
            raise Exception(str(error_res))
        to_send = [check_str, params, deps]
        return send_single_to_queue(environ, to_send, uuid)
    
    @classmethod
    def queue_action(cls, environ, action, params={}, deps=[], uuid=None):
        """
        Queue a single action, given by action function name, with given parameters
        and dependencies (both optional). Also optionally pass in a uuid, which
        will be used for the run if provided
    
        Args:
            environ (str): Foursight environment name
            action (str): action function name
            params (dict): kwargs to use for action. Defaults to {}
            deps (list): list of dependencies for the action. Defaults to []
            uuid (str): optional uuid to pass to the run. Defaults to None
    
        Returns:
            str: uuid of the queued action (from send_single_to_queue)
        """
        action_str = cls.CheckHandler.get_action_strings(action)
        if not action_str:
            error_res = {
                'status': 'error',
                'description': 'could not find action %s' % action,
                'environment': environ,
                'checks': {}
            }
            raise Exception(str(error_res))
        to_send = [action_str, params, deps]
        return send_single_to_queue(environ, to_send, uuid)
    
    @classmethod
    def send_single_to_queue(cls, environ, to_send, uuid, invoke_runner=True):
        """
        Send a single formatted check/action to SQS for given environ. Pass
        the given uuid to send_sqs_messages. Invoke a single check runner lambda
    
        Args:
            environ (str): Foursight environment name
            to_send (list): check/action entry in form [check_str, params, deps]
            uuid (str): uuid to pass to run. If None, generate a new uuid
            invoke_runner (bool): If True, invoke a check_runner lambda
    
        Returns:
            str: uuid of the queued run (from send_single_to_queue)
        """
        queue = get_sqs_queue()
        queued_uuid = send_sqs_messages(queue, environ, [to_send], uuid=uuid)
        # kick off a single check runner lambda
        if invoke_runner is True:
            invoke_check_runner({'sqs_url': queue.url})
        return queued_uuid
    
    @classmethod
    def run_check_runner(cls, runner_input, propogate=True):
        """
        Run logic for a check runner. Used to run checks and actions.
    
        runner_input should be a dict containing one
        key: sqs_url that corresponds to the aws url for the queue.
        This function attempts to recieve one message from the standard SQS queue
        using long polling, checks the run dependencies for that check/action, and then
        will run the check.
    
        If the run was a check and the 'queue_action' parameter is set, along with
        check.action and check.allow_action, will attempt to queue the associated
        action.
    
        If dependencies are not met, the check is not run and
        the run info is put back in the queue. Otherwise, the message is deleted
        from the queue.
    
        If there are no messages left (should always be true when nothing is
        recieved from sqs with long polling), then exit and do not propogate another
        check runner. Otherwise, initiate another check_runner to continue the process.
    
        Args:
            runner_input (dict): body of SQS message
            propogate (bool): if True (default), invoke another check runner lambda
    
        Returns:
            dict: run result if something was run, else None
        """
        sqs_url = runner_input.get('sqs_url')
        if not sqs_url:
            return
        client = boto3.client('sqs')
        response = client.receive_message(
            QueueUrl=sqs_url,
            AttributeNames=['MessageGroupId'],
            MaxNumberOfMessages=1,
            VisibilityTimeout=300,
            WaitTimeSeconds=10
        )
        message = response.get('Messages', [{}])[0]
        body = message.get('Body')
        receipt = message.get('ReceiptHandle')
        if not body or not receipt:
            # if no messages recieved in 10 seconds of long polling, terminate
            return None
        check_list = json.loads(body)
        if not isinstance(check_list, list) or len(check_list) != 5:
            # if not a valid check str, remove the item from the SQS
            delete_message_and_propogate(runner_input, receipt, propogate=propogate)
            return None
        [run_env, run_uuid, run_name, run_kwargs, run_deps] = check_list
        # find information from s3 about completed checks in this run
        # actual id stored in s3 has key: <run_uuid>/<run_name>
        if run_deps and isinstance(run_deps, list):
            already_run = collect_run_info(run_uuid)
            deps_w_uuid = ['/'.join([run_uuid, dep]) for dep in run_deps]
            finished_dependencies = set(deps_w_uuid).issubset(already_run)
            if not finished_dependencies:
                print('-RUN-> Not ready for: %s' % (run_name))
        else:
            finished_dependencies = True
        connection = cls.init_connection(run_env)
        if finished_dependencies:
            # add the run uuid as the uuid to kwargs so that checks will coordinate
            if 'uuid' not in run_kwargs:
                run_kwargs['uuid'] = run_uuid
            run_kwargs['_run_info'] = {'run_id': run_uuid, 'receipt': receipt, 'sqs_url': sqs_url}
            # if this is an action, ensure we have not already written an action record
            if 'check_name' in run_kwargs and 'called_by' in run_kwargs:
                rec_key = '/'.join([run_kwargs['check_name'], 'action_records', run_kwargs['called_by']])
                found_rec = connection.get_object(rec_key)
                if found_rec is not None:
                    # the action record has been written. Abort and propogate
                    print('-RUN-> Found existing action record: %s. Skipping' % rec_key)
                    delete_message_and_propogate(runner_input, receipt, propogate=propogate)
                    return None
                else:
                    # make a action record before running the action
                    # action name is the second part of run_name
                    act_name = run_name.split('/')[-1]
                    rec_body = ''.join([act_name, '/', run_uuid, '.json'])
                    connection.put_object(rec_key, rec_body)
                    print('-RUN-> Wrote action record: %s' % rec_key)
            run_result = cls.CheckHandler.run_check_or_action(connection, run_name, run_kwargs)
            print('-RUN-> RESULT:  %s (uuid)' % str(run_result.get('uuid')))
            # invoke action if running a check and kwargs['queue_action'] matches stage
            stage = cls.Config.get_stage_info()['stage']
            if run_result['type'] == 'check' and run_result['kwargs']['queue_action'] == stage:
                # must also have check.action and check.allow_action set
                if run_result['allow_action'] and run_result['action']:
                    action_params = {'check_name': run_result['name'],
                                     'called_by': run_result['kwargs']['uuid']}
                    try:
                        queue_action(run_env, run_result['action'],
                                     params=action_params, uuid=run_uuid)
                    except Exception as exc:
                        print('-RUN-> Could not queue action %s on stage %s with kwargs: %s. Error: %s'
                              % (run_result['action'], stage, action_params, str(exc)))
                    else:
                        print('-RUN-> Queued action %s on stage %s with kwargs: %s'
                              % (run_result['action'], stage, action_params))
            print('-RUN-> Finished: %s' % (run_name))
            delete_message_and_propogate(runner_input, receipt, propogate=propogate)
            return run_result
        else:
            print('-RUN-> Recovered: %s' % (run_name))
            recover_message_and_propogate(runner_input, receipt, propogate=propogate)
            return None
    
    @classmethod
    def collect_run_info(cls, run_uuid):
        """
        Returns a set of run checks under this run uuid
        """
        s3_connection = S3Connection(cls.prefix + '-runs')
        run_prefix = ''.join([run_uuid, '/'])
        complete = s3_connection.list_all_keys_w_prefix(run_prefix)
        # eliminate duplicates
        return set(complete)



from __future__ import print_function, unicode_literals
from datetime import datetime
import boto3
import json


def invoke_check_runner(runner_input):
    """
    Simple function to invoke the next check_runner lambda with runner_input
    (dict containing {'sqs_url': <str>})
    """
    client = boto3.client('lambda')
    # InvocationType='Event' makes asynchronous
    # try/except while async invokes are problematic
    try:
        response = client.invoke(
            FunctionName=get_stage_info()['runner_name'],
            InvocationType='Event',
            Payload=json.dumps(runner_input)
        )
    except:
        response = client.invoke(
            FunctionName=get_stage_info()['runner_name'],
            Payload=json.dumps(runner_input)
        )
    return response


def delete_message_and_propogate(runner_input, receipt, propogate=True):
    """
    Delete the message with given receipt from sqs queue and invoke the next
    lambda runner.

    Args:
        runner_input (dict): runner info, should minimally have 'sqs_url'
        receipt (str): SQS message receipt
        propogate (bool): if True (default), invoke another check runner lambda

    Returns:
        None
    """
    sqs_url = runner_input.get('sqs_url')
    if not sqs_url or not receipt:
        return
    client = boto3.client('sqs')
    client.delete_message(
        QueueUrl=sqs_url,
        ReceiptHandle=receipt
    )
    if propogate is True:
        invoke_check_runner(runner_input)


def recover_message_and_propogate(runner_input, receipt, propogate=True):
    """
    Recover the message with given receipt to sqs queue and invoke the next
    lambda runner.

    Changing message VisibilityTimeout to 15 seconds means the message will be
    available to the queue in that much time. This is a slight lag to allow
    dependencies to process.
    NOTE: VisibilityTimeout should be less than WaitTimeSeconds in run_check_runner

    Args:
        runner_input (dict): runner info, should minimally have 'sqs_url'
        receipt (str): SQS message receipt
        propogate (bool): if True (default), invoke another check runner lambda

    Returns:
        None
    """
    sqs_url = runner_input.get('sqs_url')
    if not sqs_url or not receipt:
        return
    client = boto3.client('sqs')
    client.change_message_visibility(
        QueueUrl=sqs_url,
        ReceiptHandle=receipt,
        VisibilityTimeout=15
    )
    if propogate is True:
        invoke_check_runner(runner_input)


def get_sqs_queue():
    """
    Returns boto3 sqs resource
    """
    queue_name = get_stage_info()['queue_name']
    sqs = boto3.resource('sqs')
    try:
        queue = sqs.get_queue_by_name(QueueName=queue_name)
    except:
        queue = sqs.create_queue(
            QueueName=queue_name,
            Attributes={
                'VisibilityTimeout': '300',
                'MessageRetentionPeriod': '3600'
            }
        )
    return queue


def send_sqs_messages(queue, environ, check_vals, uuid=None):
    """
    Send messages to SQS queue. Check_vals are entries within a check_group.
    Optionally, provide a uuid that will be queued as the uuid for the run; if
    not provided, datetime.utcnow is used

    Args:
        queue: boto3 sqs resource (from get_sqs_queue)
        environ (str): foursight environment name
        check_vals (list): list of formatted check vals, like those from
            check_utils.CheckHandler.get_check_schedule
        uuid (str): optional string uuid

    Returns:
        str: uuid of queued messages
    """
    # uuid used as the MessageGroupId
    if not uuid:
        uuid = datetime.utcnow().isoformat()
    # append environ and uuid as first elements to all check_vals
    proc_vals = [[environ, uuid] + val for val in check_vals]
    for val in proc_vals:
        response = queue.send_message(MessageBody=json.dumps(val))
    return uuid


def get_sqs_attributes(sqs_url):
    """
    Returns a dict of the desired attributes form the queue with given url
    """
    backup = {
        'ApproximateNumberOfMessages': 'ERROR',
        'ApproximateNumberOfMessagesNotVisible': 'ERROR'
    }
    client = boto3.client('sqs')
    try:
        result = client.get_queue_attributes(
            QueueUrl=sqs_url,
            AttributeNames=[
                'ApproximateNumberOfMessages',
                'ApproximateNumberOfMessagesNotVisible'
            ]
        )
    except:
        return backup
    return result.get('Attributes', backup)



from dcicutils import ff_utils, s3Utils
from datetime import datetime
from operator import itemgetter
from . import wfrset_cgap_utils
import json
lambda_limit = wfrset_cgap_utils.lambda_limit
# use wf_dict in workflow version check to make sure latest version and workflow uuid matches
wf_dict = wfrset_cgap_utils.wf_dict
# check at the end
# check extract_file_info has 4 arguments

# wfr_name, accepted versions, expected run time
workflow_details = {
    "md5": {
        "run_time": 12,
        "accepted_versions": ["0.0.4", "0.2.6"]
    },
    "fastqc-0-11-4-1": {
        "run_time": 50,
        "accepted_versions": ["0.2.0"]
    },
    "fastqc": {
        "run_time": 50,
        "accepted_versions": ["v1", "v2"]
    },
    "workflow_bwa-mem_no_unzip-check": {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_add-readgroups-check": {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_merge-bam-check": {
        "run_time": 12,
        "accepted_versions": ["v9", "v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_picard-MarkDuplicates-check": {
        "run_time": 12,
        "accepted_versions": ["v9", "v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_sort-bam-check": {
        "run_time": 12,
        "accepted_versions": ["v9", "v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_gatk-BaseRecalibrator": {
        "run_time": 12,
        "accepted_versions": ["v9", "v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_gatk-ApplyBQSR-check": {
        "run_time": 12,
        "accepted_versions": ["v9", "v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    # defunct step 8
    "workflow_index-sorted-bam": {
        "run_time": 12,
        "accepted_versions": ["v9"]
    },
    "workflow_granite-mpileupCounts": {
        "run_time": 12,
        "accepted_versions": ["v14", "v15", "v16", "v17"]
    },
    # new step 8
    'workflow_gatk-HaplotypeCaller': {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    # step 9
    'workflow_granite-mpileupCounts': {
        "run_time": 12,
        "accepted_versions": ["v14", "v15", "v16", "v17"]
    },
    # step 10
    'cgap-bamqc': {
        "run_time": 12,
        "accepted_versions": ["v2", "v3"]
    },
    # # PART II
    # part II step 1
    'workflow_gatk-CombineGVCFs': {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    # part II step 2
    'workflow_gatk-GenotypeGVCFs-check': {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13", "v15", "v16", "v17"]
    },
    # part III step 3
    'workflow_gatk-VQSR-check': {
        "run_time": 12,
        "accepted_versions": ["v10", "v11", "v12", "v13"]
    },
    "workflow_qcboard-bam": {
        "run_time": 12,
        "accepted_versions": ["v3"]
    },
    "workflow_cram2fastq": {
        "run_time": 12,
        "accepted_versions": ["v12", "v13", "v15", "v16", "v17"]
    },
    "workflow_cram2bam-check": {
        "run_time": 12,
        "accepted_versions": ["v13", "v15", "v16", "v17"]
    },
    "workflow_vep-parallel": {
        "run_time": 12,
        "accepted_versions": ["v17"]
    },
    "workflow_mutanno-micro-annot-check": {
        "run_time": 12,
        "accepted_versions": ["v17"]
    },
    # Part III
    "workflow_granite-rckTar": {
        "run_time": 12,
        "accepted_versions": ["v16", "v17"]
    },
    "workflow_granite-filtering-check": {
        "run_time": 12,
        "accepted_versions": ["v16", "v17"]
    },
    "workflow_granite-novoCaller-rck-check": {
        "run_time": 12,
        "accepted_versions": ["v16", "v17"]
    },
    "workflow_granite-comHet-check": {
        "run_time": 12,
        "accepted_versions": ["v16", "v17"]
    },
    "workflow_mutanno-annot-check": {
        "run_time": 12,
        "accepted_versions": ["v17"]
    },
    "bamsnap": {
        "run_time": 12,
        "accepted_versions": ["v17"]
    },
    "workflow_granite-qcVCF": {
        "run_time": 12,
        "accepted_versions": ["v2", "v3"]
    },
    "workflow_peddy": {
        "run_time": 12,
        "accepted_versions": ["v3"]
    }
}


# Reference Files (should be @id)
bwa_index = {'human': '/files-reference/GAPFI4U1HXIY/'}


def remove_parents_without_sample(samples_pedigree):
    individuals = [i['individual'] for i in samples_pedigree]
    for a_member in samples_pedigree:
        parents = a_member['parents']
        new_parents = [i for i in parents if i in individuals]
        a_member['parents'] = new_parents
    return samples_pedigree


def analyze_pedigree(samples_pedigree_json, all_samples):
    """extract pedigree for qc for trio (father/mother/proband) or proband
    - input samples accession list
    - qc pedigree
    - run mode (trio or proband_only)
    - error (expected empty)
    """
    # remove parent ids that are not in the sample_pedigree as individual
    samples_pedigree = remove_parents_without_sample(samples_pedigree_json)
    # Define run Mode: Trio or Proband only
    # order samples by father/mother/proband
    run_mode = ""
    error = ""
    input_samples = []
    qc_pedigree = []  # used by vcfqc
    # trio analysis
    if len(all_samples) > 2:
        for member in ['father', 'mother', 'proband']:
            sample_info = [i for i in samples_pedigree if i.get('relationship') == member]
            if not sample_info:
                break
            member_qc_pedigree = {
                'gender': sample_info[0].get('sex', ''),
                'individual': sample_info[0].get('individual', ''),
                'parents': sample_info[0].get('parents', []),
                'sample_name': sample_info[0].get('sample_name', '')
                }
            qc_pedigree.append(member_qc_pedigree)
            input_samples.append(sample_info[0]['sample_accession'])
        if len(input_samples) != 3:
            error = 'does not have mother father proband info'
            return "", "", "", error
        run_mode = 'trio'
    # if there are only 2 or less members, go for proband only
    else:
        sample_info = [i for i in samples_pedigree if i.get('relationship') == 'proband']
        if not sample_info:
            error = 'does not have proband info'
            return "", "", "", error
        input_samples.append(sample_info[0]['sample_accession'])
        member_qc_pedigree = {
            'gender': sample_info[0].get('sex', ''),
            'individual': sample_info[0].get('individual', ''),
            'parents': sample_info[0].get('parents', []),
            'sample_name': sample_info[0].get('sample_name', '')
            }
        qc_pedigree.append(member_qc_pedigree)
        run_mode = 'proband_only'
    # remove parents from mother and father (temporary fix until vcfqc V4 is in production)
    qc_pedigree = remove_parents_without_sample(qc_pedigree)
    return input_samples, qc_pedigree, run_mode, error


def get_bamsnap_parameters(samples_pedigree, all_samples):
    """collect bam @ids and titles for all samples in the sample_procesing item
    used by bamsnap
    start with proband, and continue with close relatives
    - args: samples and samples_pedigree fields of sample_processing
    - returns: 2 lists, collected bams and titles
    """

    def get_summary(a_role_info, all_samples):
        """given a samples pedigree item and all samples information (from samples pedigree)
        return a summary for bamsnap"""
        relation = a_role_info['relationship']
        sample_acc = a_role_info['sample_accession']
        sample_name = a_role_info.get('sample_name', '')
        sample_title = "{} ({})".format(sample_name, relation)
        # get sample info from samples field
        sample_info = [i for i in all_samples if sample_acc in i['@id']][0]
        bams = [i['@id'] for i in sample_info['processed_files'] if i['display_title'].endswith('bam')]
        if not bams:
            raise ValueError('can not locate bam file on sample {} to be used by bamsnap'.format(sample_acc))
        bam = bams[0]
        return {'bam': bam, 'title': sample_title, 'accession': sample_acc}

    # remove parent ids that are not in the sample_pedigree as individual
    samples_pedigree = remove_parents_without_sample(samples_pedigree)
    summary = []

    for member in ['proband', 'mother', 'father', 'brother', 'sister', 'sibling', 'half-brother', 'half-sister', 'half-sibling']:
        # if multiple relationships of same type are in the family they will have enumeration in the relationships
        # ie brother, brother II
        role_pedigree_infos = [i for i in samples_pedigree if i.get('relationship', '').split(' ')[0] == member]
        if not role_pedigree_infos:
            continue
        for a_role_pedigree in role_pedigree_infos:
            sample_summary = get_summary(a_role_pedigree, all_samples)
            summary.append(sample_summary)
    # for family members not listed on primary relations, get same information and continue appending
    all_sample_accs = [i['@id'].split('/')[2] for i in all_samples]
    seen_sample_accs = [i['accession'] for i in summary]
    remaining_samples = [i for i in all_sample_accs if i not in seen_sample_accs]

    for a_sample_acc in remaining_samples:
        a_role_pedigree = [i for i in samples_pedigree if i['sample_accession'] == a_sample_acc][0]
        sample_summary = get_summary(a_role_pedigree, all_samples)
        summary.append(sample_summary)
    bams = [i['bam'] for i in summary]
    titles = [i['title'] for i in summary]
    return bams, titles


def check_latest_workflow_version(workflows):
    """Some sanity checks for workflow versions
    expectations:
     - All workflows that we are currently active should be listed both on
       cgap_utils.py (workflow_details) and wfrset_cgap_utils.py (wf_dict)
     - The lastest workflow version on foursight (workflow_details) should be carried by the
       latest released workflow item on the data portal.
       If a new version is released on the portal, we need it to be on foursight too, if not stop the check.
       If a new version is decleared on foursight, it should be released on the portal, if not stop the check.
    """
    errors = []
    for a_wf in workflows:
        wf_name = a_wf['app_name']
        # make sure the workflow is in our control list on cgap_utils.py
        if wf_name not in workflow_details:
            continue
        # make sure the workflow is in our settings list on wfrset_cgap_utils.py
        if wf_name not in [i['app_name'] for i in wf_dict]:
            continue
        wf_info = workflow_details[wf_name]
        versions = wf_info['accepted_versions']
        # latest version should be the last one on the list
        last_version = versions[-1]
        # sometimes there are 2 or more workflows with same app name
        # and the old one might not have the latest version
        # look for all wfs with same name and make sure the latest version is on one
        same_wf_name_workflows = [i for i in workflows if i['app_name'] == wf_name]
        all_wf_versions = [i.get('app_version', '') for i in same_wf_name_workflows]
        # make sure the latest is also on one of the wfrs
        if last_version not in all_wf_versions:
            err = '{} version {} is not on any wf app_version)'.format(wf_name, last_version)
            errors.append(err)
            continue
        # fist item on same_wf_name_workflows should be the latest released workflow, check if we added that to foursight
        last_wf_version_on_portal = all_wf_versions[0]
        if last_wf_version_on_portal not in all_wf_versions:
            err = '{} version {} is not decleared on foursight)'.format(wf_name, last_wf_version_on_portal)
            errors.append(err)
            continue
        # check if the lastest version workflow uuids is correct on wfr_dict (wfrset_cgap_utils.py)
        latest_workflow_uuid = [i['uuid'] for i in same_wf_name_workflows if i['app_version'] == last_version][0]
        wf_dict_item = [i['workflow_uuid'] for i in wf_dict if i['app_name'] == wf_name][0]
        if latest_workflow_uuid != wf_dict_item:
            err = '{} item on wf_dict does not have the latest workflow uuid'.format(wf_name)
            errors.append(err)
            continue
    # return unique errors
    return list(set(errors))


def check_qcs_on_files(file_meta, all_qcs):
    """Go over qc related fields, and check for overall quality score."""
    def check_qc(file_accession, resp, failed_qcs_list):
        """format errors and return a errors list."""
        quality_score = resp.get('overall_quality_status', '')
        if quality_score.upper() != 'PASS':
            failed_qcs_list.append([file_accession, resp['display_title'], resp['uuid']])
        return failed_qcs_list

    failed_qcs = []
    if not file_meta.get('quality_metric'):
        return
    qc_result = [i for i in all_qcs if i['@id'] == file_meta['quality_metric']['@id']][0]
    if qc_result['display_title'].startswith('QualityMetricQclist'):
        if not qc_result.get('qc_list'):
            return
        for qc in qc_result['qc_list']:
            qc_resp = [i for i in all_qcs if i['@id'] == qc['value']['@id']][0]
            failed_qcs = check_qc(file_meta['accession'], qc_resp, failed_qcs)
    else:
        failed_qcs = check_qc(file_meta['accession'], qc_result, failed_qcs)
    return failed_qcs


def order_input_dictionary(input_file_dict):
    """Keep the order of file_arg keys in dictionary, but if the value is a list of files
    order them to be able to compare them"""
    ordered_input = {}
    for an_input_arg in input_file_dict:
        if isinstance(input_file_dict[an_input_arg], (list, tuple)):
            ordered_input[an_input_arg] = sorted(input_file_dict[an_input_arg])
        else:
            ordered_input[an_input_arg] = input_file_dict[an_input_arg]
    return ordered_input


def collect_input_files_from_input_dictionary(input_file_dict):
    """Collect all @ids from the input file dictionary"""
    all_inputs = []
    for an_input_arg in input_file_dict:
        if an_input_arg == 'additional_file_parameters':
            continue
        if isinstance(input_file_dict[an_input_arg], (list, tuple)):
            all_inputs.extend(input_file_dict[an_input_arg])
        else:
            all_inputs.append(input_file_dict[an_input_arg])
    return all_inputs


def collect_inputs_from_workflow_run(wfr_resp):
    """Given a wfr item in embedded frame, collect input files as a list of @ids"""
    input_files = []
    inputs = wfr_resp['input_files']
    for an_input in inputs:
        if an_input.get('value'):
            input_files.append(an_input['value']['@id'])
    return input_files


def remove_duplicate_need_runs(need_runs_dictionary):
    """In rare cases, the same run can be triggered by two different sample_processings
    example is a quad analyzed for 2 different probands. In this case you want a single
    combineGVCF step, but 2 sample_processings will be generated trying to run same job,
    identify and remove duplicates, used by PartII
    input_structure:
    [{"sample_processing_id": [
            # list of items per run
            ["step_id",  # name for the run
             ["workflow_app_name", "organism", "additional_parameters"],  # run settings
             {"input_arg1": ["file_id1", "file_id2"], "input_arg2": "file_id3"},  # run input files
             "GAPFI1RGV5US_GAPFIOASB96R_GAPFIVPH2MVG_GAPFICMTHU2P"  # all input files going in the run
             ]]}]
    """
    # keep a track of sorted runs
    sorted_runs = []
    unique_needs_runs_dictionary = []

    for an_item in need_runs_dictionary:
        for an_sp_id in an_item:
            keep_runs = []
            all_runs = an_item[an_sp_id]
            for a_run in all_runs:
                # sort input files
                run_setttings = a_run[1]
                input_files = a_run[2]
                ordered_input = order_input_dictionary(input_files)
                # get elements that should be unique and store as string
                run_info = str(run_setttings) + str(ordered_input)
                if run_info in sorted_runs:
                    continue
                else:
                    keep_runs.append(a_run)
                    sorted_runs.append(run_info)
            if keep_runs:
                unique_needs_runs_dictionary.append({an_sp_id: keep_runs})
    return unique_needs_runs_dictionary


def filter_wfrs_with_input_and_tag(all_wfr_items, app_name, input_file_dict, tag, match_all_input=True):
    """given an input file dictionary and list of workflow_run items, filter wfrs
    for input files that match the input file dictionary. If a filter tag is given
    also filter for workflow_runs that have the given filter_tag in tags field
    -- args
    all_wfr_items: all workflow run items collected from ES
    step_name: workflow_app_name
    input_file_dict: all input files
    tag: if filter should look for a tag (ie sample_processing uuid) on wfr
    match_all_input: (Bool) True by default. Looks for exact match between input file dict and the workflowrun
    # TODO: for combine GVCF, we need to implement the False case, where we utilize the partial match
    """
    # filter workflows for workflow_name, the ones with same input files, and if exist, tags
    # filtering with input - important for steps like combine gVCF there same input file
    #                        might have multiple runs of same type with different input
    #                        combinations, coming from different sample_procesing items.
    # filtering with tag - for some steps, even if the input files are the same,
    #                      you need to run different versions for different sample processing items
    #                      (ie 2 quads made up of the same samples with different probands.)

    # filter for app_name
    wfrs_with_app_name = [i for i in all_wfr_items if i['display_title'].startswith(app_name)]
    # filter for tag
    if tag:
        wfrs_with_tag = [i for i in wfrs_with_app_name if tag in i.get('tags', [])]
    else:
        wfrs_with_tag = wfrs_with_app_name
    # filter for input files
    # collect input files
    input_files = collect_input_files_from_input_dictionary(input_file_dict)
    # check for workflows with same inputs
    filtered_wfrs = []
    for a_wfr in wfrs_with_tag:
        # get all input files from the workflow item
        wfr_inputs = collect_inputs_from_workflow_run(a_wfr)
        # expectation is exact match of input files, if so pass the filter
        if match_all_input:
            if sorted(wfr_inputs) == sorted(input_files):
                filtered_wfrs.append(a_wfr)
        # if input files are contained by the wfr input files, pass the filter
        else:
            if all(file in wfr_inputs for file in input_files):
                filtered_wfrs.append(a_wfr)
    return filtered_wfrs


def check_input_structure_at_id(input_file_dict):
    """Check all input file strings and make sure they are @id format."""
    all_inputs = []
    error = ''
    for an_input_arg in input_file_dict:
        # skip the parameter key
        if an_input_arg == 'additional_file_parameters':
            continue
        inputs = input_file_dict[an_input_arg]
        # if the input is array
        if isinstance(inputs, list) or isinstance(inputs, tuple):
            all_inputs.extend(inputs)
        else:
            all_inputs.append(inputs)
    for an_input in all_inputs:
        if an_input.count('/') != 3:
            error += an_input + ' '
    if error:
        error += 'files are not @ids, foursight needs update'
        return [error, ]
    else:
        return []


def stepper(library, keep,
            step_tag, new_step_input_file,
            input_file_dict,  new_step_name, new_step_output_arg,
            additional_input=None, organism='human', no_output=False, tag=''):
    """This functions packs the core of wfr check, for a given workflow and set of
    input files, it will return the status of process on these files.
    It will also check for failed qcs on input files.
    - args
      -library:   dictionary with keys files/wfrs/qcs that contain all related items
      -keep:      tracking run progress with keys running/problematic_run/missing_run
      -step_tag:  informative summary used in the output (ie step name + input file accession)
      -new_step_input_file:  files to check for qc and get attribution from
      -input_file_dict:      all files and arguments used in the wfr
      -new_step_name:        workflow app_name
      -new_step_output_arg:  can be str or list, will return str or list of @id for output files with given argument(s)
      -additional_input:     overwrite or add to final json, example {"parameters": {'key': 'value'}, "config": {"key": "value"}}
      -organism:    by default human, used for adding genome assembly
      -no_output:   default False, should be True for workflows that don't produce an output file (ie QC runs)
      -tag:         if used, filter workflow_run items for the ones which have this as a tag, or add the tag when creating runs
                    (used for distinguishing sample_processing specific steps even when input files are the same (ie micro annotation))
    - returns
        - keep : same as input, with addition from this check
        - step_status : a short summary of this functions result (complete, running, no complete run)
    """
    if not additional_input:
        additional_input = {}
    step_output = ''
    # unpack library
    all_files = library['files']
    all_wfrs = library['wfrs']
    all_qcs = library['qcs']
    # unpack keep
    running = keep['running']
    problematic_run = keep['problematic_run']
    missing_run = keep['missing_run']

    # make sure input files are @ids, if not foursight needs an update, report it as error
    at_id_errors = check_input_structure_at_id(input_file_dict)

    # Lets get the repoinse from one of the input files that will be used in this step
    # if it is a list take the first item, if not use it as is
    # new_step_input_file must be the @id
    # also check for qc status
    qc_errors = []
    if isinstance(new_step_input_file, list) or isinstance(new_step_input_file, tuple):
        for an_input in new_step_input_file:
            input_resp = [i for i in all_files if i['@id'] == an_input][0]
            errors = check_qcs_on_files(input_resp, all_qcs)
            if errors:
                qc_errors.extend(errors)
        name_tag = '_'.join([i.split('/')[2] for i in new_step_input_file])
    else:
        input_resp = [i for i in all_files if i['@id'] == new_step_input_file][0]
        errors = check_qcs_on_files(input_resp, all_qcs)
        if errors:
            qc_errors.extend(errors)
        name_tag = new_step_input_file.split('/')[2]
    # if there are @id errors return it
    if at_id_errors:
        problematic_run.append([step_tag + ' foursight error', at_id_errors])
        step_status = "no complete run, foursight error"
    # if there are qc errors, return with qc qc_errors
    elif qc_errors:
        problematic_run.append([step_tag + ' input file qc error', qc_errors])
        step_status = "no complete run, qc error"
    # if no qc problem, go on with the run check
    else:
        # filter workflows for the ones with same input files, and if exist, tags
        # filtering with input - important for steps like combine gVCF there same input file
        #                        might have multiple runs of same type with different input
        #                        combinations, coming from different sample_procesing items.
        # filtering with tag - for some steps, even if the input files are the same,
        #                      you need to run different versions for different sample processing items
        #                      (ie 2 quads made up of the same samples with different probands.)
        filtered_wfrs = filter_wfrs_with_input_and_tag(all_wfrs, new_step_name, input_file_dict, tag=tag)

        if no_output:
            step_result = get_wfr_out(input_resp, new_step_name, all_wfrs=filtered_wfrs, md_qc=True)
        else:
            step_result = get_wfr_out(input_resp, new_step_name, all_wfrs=filtered_wfrs)
        step_status = step_result['status']
        # if successful
        input_file_accession = input_resp['accession']
        if step_status == 'complete':
            if new_step_output_arg:
                if isinstance(new_step_output_arg, list):
                    output_list = []
                    for an_output_arg in new_step_output_arg:
                        output_list.append(step_result[an_output_arg])
                    step_output = output_list
                else:
                    step_output = step_result[new_step_output_arg]
            pass
        # if still running
        elif step_status == 'running':
            running.append([step_tag, input_file_accession])
        # if run is not successful
        elif step_status.startswith("no complete run, too many"):
            problematic_run.append([step_tag, input_file_accession])
        else:
            # add missing run
            # if there is a tag, pass it to the workflow_run metadata
            if tag:
                if not additional_input.get('wfr_meta'):
                    additional_input['wfr_meta'] = {'tags': [tag, ]}
                else:
                    additional_input['wfr_meta']['tags'] = [tag, ]
            missing_run.append([step_tag, [new_step_name, organism, additional_input], input_file_dict, name_tag])

    keep['running'] = running
    keep['problematic_run'] = problematic_run
    keep['missing_run'] = missing_run
    return keep, step_status, step_output


def get_wfr_out(emb_file, wfr_name, key=None, all_wfrs='not given', versions=None, md_qc=False, run=None):
    """For a given file, fetches the status of last wfr (of wfr_name type)
    If there is a successful run, it will return the output files as a dictionary of
    argument_name:file_id, else, will return the status. Some runs, like qc and md5,
    does not have any file_format output, so they will simply return 'complete'
    Can be run in two modes
     - Search through given wfr library on all_wfrs parameter, dont give any key
     - Search on the database without a given library (all_wfrs = not given and there should be a key) used by qcs
    args:
     emb_file: embedded frame file info
     wfr_name: base name without version
     key: authorization
     all_wfrs : all releated wfrs in embedded frame
                to distinguish
     versions: acceptable versions for wfr
     md_qc: if no output file is excepted, set to True
     run: if run is still running beyond this hour limit, assume problem
    """
    # sanity checks
    # we need key if all wfrs is not supplied (it can even be empty list but needs to be provided)
    if not key and all_wfrs == 'not given':
        raise ValueError('library or key is required for get_wfr_out function')

    error_at_failed_runs = 1
    # you should provide key or all_wfrs
    # assert key or all_wfrs
    assert wfr_name in workflow_details
    # get default accepted versions if not provided
    if not versions:
        versions = workflow_details[wfr_name]['accepted_versions']
    # get default run out time
    if not run:
        run = workflow_details[wfr_name]['run_time']
    workflows = emb_file.get('workflow_run_inputs', [])
    wfr = {}
    run_status = 'did not run'
    wfrs_on_file = [i for i in workflows if i['display_title'].startswith(wfr_name)]
    # if all_wfrs is not given, get workflows from the file
    if all_wfrs == 'not given':
        my_workflows = wfrs_on_file
    # otherwise, limit the workflows to the ones from all_wfrs
    else:
        library_uuids = [i['uuid'] for i in all_wfrs]
        my_workflows = [i for i in wfrs_on_file if i['uuid'] in library_uuids]
    if not my_workflows:
        return {'status': "no workflow on file"}
    # if all_wfrs were given and there were no wfrs, it means that prefiltering did not return any
    if not key and not all_wfrs:
        return {'status': "no workflow on file"}

    for a_wfr in my_workflows:
        wfr_type, time_info = a_wfr['display_title'].split(' run ')
        wfr_type_base, wfr_version = wfr_type.strip().split(' ')
        # user submitted ones use run on insteand of run
        time_info = time_info.strip('on').strip()
        try:
            wfr_time = datetime.strptime(time_info, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            wfr_time = datetime.strptime(time_info, '%Y-%m-%d %H:%M:%S')
        a_wfr['run_hours'] = (datetime.utcnow() - wfr_time).total_seconds() / 3600
        a_wfr['run_type'] = wfr_type_base.strip()
        a_wfr['run_version'] = wfr_version.strip()
    my_workflows = [i for i in my_workflows if i['run_version'] in versions]
    if not my_workflows:
        return {'status': "no workflow in file with accepted version"}
    my_workflows = sorted(my_workflows, key=lambda k: k['run_hours'])
    same_type_wfrs = [i for i in my_workflows if i['run_type'] == wfr_name]
    last_wfr = same_type_wfrs[0]
    # get metadata for the last wfr
    if all_wfrs == 'not given':
        wfr = ff_utils.get_metadata(last_wfr['uuid'], key)
    else:
        wfr = [i for i in all_wfrs if i['uuid'] == last_wfr['uuid']][0]
    run_duration = last_wfr['run_hours']
    run_status = wfr['run_status']

    if run_status == 'complete':
        outputs = wfr.get('output_files')
        # some runs, like qc, don't have a real file output
        if md_qc:
            return {'status': 'complete'}
        # if expected output files, return a dictionary of argname:file_id
        else:
            out_files = {}
            for output in outputs:
                if output.get('format'):
                    # get the arg name
                    arg_name = output['workflow_argument_name']
                    out_files[arg_name] = output['value']['@id']
            if out_files:
                out_files['status'] = 'complete'
                return out_files
            else:
                return {'status': "no file found"}
    # if status is error
    elif run_status == 'error':
        # are there too many failed runs
        if len(same_type_wfrs) >= error_at_failed_runs:
            return {'status': "no complete run, too many errors"}

        return {'status': "no complete run, errrored"}
    # if other statuses, started running
    elif run_duration < run:
        return {'status': "running"}
    # this should be the timeout case
    else:
        if len(same_type_wfrs) >= error_at_failed_runs:
            return {'status': "no complete run, too many time-outs"}
        else:
            return {'status': "no completed run, time-out"}


def get_attribution(file_json):
    """give file response in embedded frame and extract attribution info"""
    attributions = {
        'project': file_json['project']['@id'],
        'institution': file_json['institution']['@id']
    }
    return attributions


def extract_file_info(obj_id, arg_name, additional_parameters, auth, env, rename=[]):
    """Takes file id, and creates info dict for tibanna"""
    my_s3_util = s3Utils(env=env)
    raw_bucket = my_s3_util.raw_file_bucket
    out_bucket = my_s3_util.outfile_bucket
    """Creates the formatted dictionary for files.
    """
    # start a dictionary
    template = {"workflow_argument_name": arg_name}
    if rename:
        change_from = rename[0]
        change_to = rename[1]
    # if it is list of items, change the structure
    if isinstance(obj_id, list):
        object_key = []
        uuid = []
        buckets = []
        for obj in obj_id:
            metadata = ff_utils.get_metadata(obj, key=auth)
            object_key.append(metadata['display_title'])
            uuid.append(metadata['uuid'])
            # get the bucket
            if 'FileProcessed' in metadata['@type']:
                my_bucket = out_bucket
            else:  # covers cases of FileFastq, FileReference, FileMicroscopy
                my_bucket = raw_bucket
            buckets.append(my_bucket)
        # check bucket consistency
        assert len(list(set(buckets))) == 1
        template['object_key'] = object_key
        template['uuid'] = uuid
        template['bucket_name'] = buckets[0]
        if rename:
            template['rename'] = [i.replace(change_from, change_to) for i in template['object_key']]
        if additional_parameters:
            template.update(additional_parameters)

    # if obj_id is a string
    else:
        metadata = ff_utils.get_metadata(obj_id, key=auth)
        template['object_key'] = metadata['display_title']
        template['uuid'] = metadata['uuid']
        # get the bucket
        if 'FileProcessed' in metadata['@type']:
            my_bucket = out_bucket
        else:  # covers cases of FileFastq, FileReference, FileMicroscopy
            my_bucket = raw_bucket
        template['bucket_name'] = my_bucket
        if rename:
            template['rename'] = template['object_key'].replace(change_from, change_to)
        if additional_parameters:
            template.update(additional_parameters)
    return template


def start_missing_run(run_info, auth, env):
    # arguments for finding the file with the attribution (as opposed to reference files)
    attr_keys = ['fastq1', 'fastq', 'input_pairs', 'input_bams',
                 'fastq_R1', 'input_bam', 'input_gvcf', 'cram',
                 'input_gvcfs', 'input_rcks', 'input_vcf',
                 '']
    run_settings = run_info[1]
    inputs = run_info[2]
    name_tag = run_info[3]
    # find file to use for attribution
    attr_file = ''
    for attr_key in attr_keys:
        if attr_key in inputs:
            attr_file = inputs[attr_key]
            if isinstance(attr_file, list):
                attr_file = attr_file[0]
            break
    if not attr_file:
        possible_keys = [i for i in inputs.keys() if i != 'additional_file_parameters']
        error_message = ('one of these argument names {} which carry the input file -not the references-'
                         ' should be added to att_keys dictionary on foursight cgap_utils.py function start_missing_run').format(possible_keys)
        raise ValueError(error_message)
    attributions = get_attribution(ff_utils.get_metadata(attr_file, auth))
    settings = wfrset_cgap_utils.step_settings(run_settings[0], run_settings[1], attributions, run_settings[2])
    url = run_missing_wfr(settings, inputs, name_tag, auth, env)
    return url


def run_missing_wfr(input_json, input_files_and_params, run_name, auth, env):
    all_inputs = []
    # input_files container
    input_files = {k: v for k, v in input_files_and_params.items() if k != 'additional_file_parameters'}
    # additional input file parameters
    input_file_parameters = input_files_and_params.get('additional_file_parameters', {})
    for arg, files in input_files.items():
        additional_params = input_file_parameters.get(arg, {})
        inp = extract_file_info(files, arg, additional_params, auth, env)
        all_inputs.append(inp)
    # tweak to get bg2bw working
    all_inputs = sorted(all_inputs, key=itemgetter('workflow_argument_name'))
    my_s3_util = s3Utils(env=env)
    out_bucket = my_s3_util.outfile_bucket
    # shorten long name_tags
    # they get combined with workflow name, and total should be less then 80
    # (even less since repeats need unique names)
    if len(run_name) > 30:
        run_name = run_name[:30] + '...'
    """Creates the trigger json that is used by foufront endpoint.
    """
    input_json['input_files'] = all_inputs
    input_json['output_bucket'] = out_bucket
    input_json["_tibanna"] = {
        "env": env,
        "run_type": input_json['app_name'],
        "run_id": run_name}
    # input_json['env_name'] = CGAP_ENV_WEBPROD  # e.g., 'fourfront-cgap'
    input_json['step_function_name'] = 'tibanna_zebra'
    input_json['public_postrun_json'] = True
    try:
        e = ff_utils.post_metadata(input_json, 'WorkflowRun/run', key=auth)
        url = json.loads(e['input'])['_tibanna']['url']
        return url
    except Exception as e:
        return str(e)


def check_runs_without_output(res, check, run_name, my_auth, start):
    """Common processing for checks that are running on files and not producing output files
    like qcs ones producing extra files"""
    # no successful run
    missing_run = []
    # successful run but no expected metadata change (qc or extra file)
    missing_meta_changes = []
    # still running
    running = []
    # multiple failed runs
    problems = []

    for a_file in res:
        # lambda has a time limit (300sec), kill before it is reached so we get some results
        now = datetime.utcnow()
        if (now-start).seconds > lambda_limit:
            check.brief_output.append('did not complete checking all')
            break
        file_id = a_file['accession']
        report = get_wfr_out(a_file, run_name,  key=my_auth, md_qc=True)
        if report['status'] == 'running':
            running.append(file_id)
        elif report['status'].startswith("no complete run, too many"):
            problems.append(file_id)
        elif report['status'] != 'complete':
            missing_run.append(file_id)
        # There is a successful run, but no extra_file
        elif report['status'] == 'complete':
            missing_meta_changes.append(file_id)
    if running:
        check.summary = 'Some files are running'
        check.brief_output.append(str(len(running)) + ' files are still running.')
        check.full_output['running'] = running
    if problems:
        check.summary = 'Some files have problems'
        check.brief_output.append(str(len(problems)) + ' files have multiple failed runs')
        check.full_output['problems'] = problems
        check.status = 'WARN'
    if missing_run:
        check.allow_action = True
        check.summary = 'Some files are missing runs'
        check.brief_output.append(str(len(missing_run)) + ' files lack a successful run')
        check.full_output['files_without_run'] = missing_run
        check.status = 'WARN'
    if missing_meta_changes:
        check.allow_action = True
        check.summary = 'Some files are missing runs'
        check.brief_output.append(str(len(missing_meta_changes)) + ' files have successful run but no qc/extra file')
        check.full_output['files_without_changes'] = missing_meta_changes
        check.status = 'WARN'
    check.summary = check.summary.strip()
    if not check.brief_output:
        check.brief_output = ['All Good!']
    return check


def find_fastq_info(my_sample, fastq_files, organism='human'):
    """Find fastq files from sample
    expects my_rep_set to be set response in frame object (search result)
    will check if files are paired or not, and if paired will give list of lists for sample
    if not paired, with just give list of files for sample.

    result is 2 lists
    - file [file1, file2, file3, file4]  # unpaired
      file [ [file1, file2], [file3, file4]] # paired
    - refs keys  {pairing, organism, bwa_ref, f_size}
    """
    # # TODO: re word for samples
    files = []
    refs = {}
    # check pairing for the first file, and assume all same
    paired = ""
    # check if files are FileFastq or FileProcessed
    f_type = ""
    total_f_size = 0
    sample_files = my_sample['files']
    # Assumption: Fastq files are either all FileFastq or File processed
    # File Processed ones don't have paired end information
    # Assumption: File Processed fastq files are paired end in the order they are in sample files
    types = [i['@id'].split('/')[1] for i in fastq_files]
    f_type = list(set(types))
    msg = '{} has mixed fastq files types {}'.format(my_sample['accession'], f_type)
    assert len(f_type) == 1, msg
    f_type = f_type[0]

    if f_type == 'files-processed':
        for fastq_file in sample_files:
            file_resp = [i for i in fastq_files if i['uuid'] == fastq_file['uuid']][0]
            if file_resp.get('file_size'):
                total_f_size += file_resp['file_size']
        # we are assuming that this files are processed
        # # TODO: make sure that this is encoded in the metadata
        paired = 'Yes'
        file_ids = [i['@id'] for i in sample_files]
        files = [file_ids[i:i+2] for i in range(0, len(file_ids), 2)]

    elif f_type == 'files-fastq':
        for fastq_file in sample_files:
            file_resp = [i for i in fastq_files if i['uuid'] == fastq_file['uuid']][0]
            if file_resp.get('file_size'):
                total_f_size += file_resp['file_size']
            # skip pair no 2
            if file_resp.get('paired_end') == '2':
                continue
            # check that file has a pair
            f1 = file_resp['@id']
            f2 = ""
            # assign pairing info by the first file
            if not paired:
                try:
                    relations = file_resp['related_files']
                    paired_files = [relation['file']['@id'] for relation in relations
                                    if relation['relationship_type'] == 'paired with']
                    assert len(paired_files) == 1
                    paired = "Yes"
                except:
                    paired = "No"

            if paired == 'No':
                files.append(f1)
            elif paired == 'Yes':
                relations = file_resp['related_files']
                paired_files = [relation['file']['@id'] for relation in relations
                                if relation['relationship_type'] == 'paired with']
                assert len(paired_files) == 1
                f2 = paired_files[0]
                files.append((f1, f2))
    bwa = bwa_index.get(organism)
    # chrsize = chr_size.get(organism)

    f_size = int(total_f_size / (1024 * 1024 * 1024))
    refs = {'pairing': paired,
            'organism': organism,
            'bwa_ref': bwa,
            'f_size': str(f_size)+'GB'}
    return files, refs


def start_tasks(missing_runs, patch_meta, action, my_auth, my_env, start, move_to_pc=False, runtype='hic'):
    started_runs = 0
    patched_md = 0
    action.description = ""
    action_log = {'started_runs': [], 'failed_runs': [], 'patched_meta': [], 'failed_meta': []}
    if missing_runs:
        for a_case in missing_runs:
            now = datetime.utcnow()
            acc = list(a_case.keys())[0]
            print((now-start).seconds, acc)
            if (now-start).seconds > lambda_limit:
                action.description = 'Did not complete action due to time limitations.'
                break

            for a_run in a_case[acc]:
                started_runs += 1
                url = start_missing_run(a_run, my_auth, my_env)
                log_message = acc + ' started running ' + a_run[0] + ' with ' + a_run[3]
                if url.startswith('http'):
                    action_log['started_runs'].append([log_message, url])
                else:
                    action_log['failed_runs'].append([log_message, url])
    if patch_meta:
        action_log['patched_meta'] = []
        for a_completed_info in patch_meta:
            exp_acc = a_completed_info[0]
            patch_body = a_completed_info[1]
            now = datetime.utcnow()
            if (now-start).seconds > lambda_limit:
                action.description = 'Did not complete action due to time limitations.'
                break
            patched_md += 1
            ff_utils.patch_metadata(patch_body, exp_acc, my_auth)
            action_log['patched_meta'].append(exp_acc)

    # did we complete without running into time limit
    for k in action_log:
        if action_log[k]:
            add_desc = "| {}: {} ".format(k, str(len(action_log[k])))
            action.description += add_desc

    action.output = action_log
    action.status = 'DONE'
    return action


def is_there_my_qc_metric(file_meta, qc_metric_name, my_auth):
    if not file_meta.get('quality_metric'):
        return False
    qc_results = ff_utils.get_metadata(file_meta['quality_metric']['uuid'], key=my_auth)
    if qc_results['display_title'].startswith('QualityMetricQclist'):
        if not qc_results.get('qc_list'):
            return False
        for qc in qc_results['qc_list']:
            if qc_metric_name not in qc['value']['display_title']:
                return False
    else:
        if qc_metric_name not in qc_results['display_title']:
            return False
    return True


def fetch_wfr_associated(wfr_info):
    """Given wfr embedded frame, find associated output files and qcs"""
    wfr_as_list = []
    wfr_as_list.append(wfr_info['uuid'])
    if wfr_info.get('output_files'):
        for o in wfr_info['output_files']:
            if o.get('value'):
                wfr_as_list.append(o['value']['uuid'])
            elif o.get('value_qc'):
                wfr_as_list.append(o['value_qc']['uuid'])
    if wfr_info.get('output_quality_metrics'):
        for qc in wfr_info['output_quality_metrics']:
            if qc.get('value'):
                wfr_as_list.append(qc['value']['uuid'])
    if wfr_info.get('quality_metric'):
        wfr_as_list.append(wfr_info['quality_metric']['uuid'])
    return list(set(wfr_as_list))


def string_to_list(string):
    "Given a string that is either comma separated values, or a python list, parse to list"
    for a_sep in "'\":[] ":
        values = string.replace(a_sep, ",")
    values = [i.strip() for i in values.split(',') if i]
    return values
